from typing import Any, Type
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from ._classes.sticky_note import StickyNote
from .classes import Enum
from .classes import Project
from .classes import Reference
from .classes import Table
from .classes import TableGroup
from .exceptions import DatabaseValidationError
from .renderer.base import BaseRenderer
from .renderer.dbml.default.renderer import DefaultDBMLRenderer
from .renderer.sql.default import DefaultSQLRenderer


class Database:
    def __init__(
        self,
        sql_renderer: Type[BaseRenderer] = DefaultSQLRenderer,
        dbml_renderer: Type[BaseRenderer] = DefaultDBMLRenderer,
        allow_properties: bool = False
    ) -> None:
        self.sql_renderer = sql_renderer
        self.dbml_renderer = dbml_renderer
        self.tables: List['Table'] = []
        self.table_dict: Dict[str, 'Table'] = {}
        self.refs: List['Reference'] = []
        self.enums: List['Enum'] = []
        self.table_groups: List['TableGroup'] = []
        self.sticky_notes: List['StickyNote'] = []
        self.project: Optional['Project'] = None
        self.allow_properties = allow_properties

    def __repr__(self) -> str:
        return f"<Database>"

    def __getitem__(self, k: Union[int, str]) -> Table:
        if isinstance(k, int):
            return self.tables[k]
        elif isinstance(k, str):
            return self.table_dict[k]
        else:
            raise TypeError('indeces must be str or int')

    def __iter__(self):
        return iter(self.tables)

    def _set_database(self, obj: Any) -> None:
        obj.database = self

    def _unset_database(self, obj: Any) -> None:
        obj.database = None

    def add(self, obj: Any) -> Any:
        if isinstance(obj, Table):
            return self.add_table(obj)
        elif isinstance(obj, Reference):
            return self.add_reference(obj)
        elif isinstance(obj, Enum):
            return self.add_enum(obj)
        elif isinstance(obj, TableGroup):
            return self.add_table_group(obj)
        elif isinstance(obj, Project):
            return self.add_project(obj)
        elif isinstance(obj, StickyNote):
            return self.add_sticky_note(obj)
        else:
            raise DatabaseValidationError(f'Unsupported type {type(obj)}.')

    def add_table(self, obj: Table) -> Table:
        if obj in self.tables:
            raise DatabaseValidationError(f'{obj} is already in the database.')
        if obj.full_name in self.table_dict:
            raise DatabaseValidationError(f'Table {obj.full_name} is already in the database.')
        if obj.alias and obj.alias in self.table_dict:
            raise DatabaseValidationError(f'Table {obj.alias} is already in the database.')

        self._set_database(obj)

        self.tables.append(obj)
        self.table_dict[obj.full_name] = obj
        if obj.alias:
            self.table_dict[obj.alias] = obj
        return obj

    def add_reference(self, obj: Reference):
        for col in (*obj.col1, *obj.col2):
            if col.table and col.table.database == self:
                break
        else:
            raise DatabaseValidationError(
                'Cannot add reference. At least one of the referenced tables'
                ' should belong to this database'
            )
        if obj in self.refs:
            raise DatabaseValidationError(f'{obj} is already in the database.')

        self._set_database(obj)
        self.refs.append(obj)
        return obj

    def add_enum(self, obj: Enum) -> Enum:
        if obj in self.enums:
            raise DatabaseValidationError(f'{obj} is already in the database.')
        for enum in self.enums:
            if enum.name == obj.name and enum.schema == obj.schema:
                raise DatabaseValidationError(f'Enum {obj.schema}.{obj.name} is already in the database.')

        self._set_database(obj)
        self.enums.append(obj)
        return obj

    def add_sticky_note(self, obj: StickyNote) -> StickyNote:
        self._set_database(obj)
        self.sticky_notes.append(obj)
        return obj

    def add_table_group(self, obj: TableGroup) -> TableGroup:
        if obj in self.table_groups:
            raise DatabaseValidationError(f'{obj} is already in the database.')
        for table_group in self.table_groups:
            if table_group.name == obj.name:
                raise DatabaseValidationError(f'TableGroup {obj.name} is already in the database.')

        self._set_database(obj)
        self.table_groups.append(obj)
        return obj

    def add_project(self, obj: Project) -> Project:
        if self.project:
            self.delete_project()
        self._set_database(obj)
        self.project = obj
        return obj

    def delete(self, obj: Any) -> Any:
        if isinstance(obj, Table):
            return self.delete_table(obj)
        elif isinstance(obj, Reference):
            return self.delete_reference(obj)
        elif isinstance(obj, Enum):
            return self.delete_enum(obj)
        elif isinstance(obj, TableGroup):
            return self.delete_table_group(obj)
        elif isinstance(obj, Project):
            return self.delete_project()
        else:
            raise DatabaseValidationError(f'Unsupported type {type(obj)}.')

    def delete_table(self, obj: Table) -> Table:
        try:
            index = self.tables.index(obj)
        except ValueError:
            raise DatabaseValidationError(f'{obj} is not in the database.')
        self._unset_database(self.tables.pop(index))
        result = self.table_dict.pop(obj.full_name)
        if obj.alias:
            self.table_dict.pop(obj.alias)
        return result

    def delete_reference(self, obj: Reference) -> Reference:
        try:
            index = self.refs.index(obj)
        except ValueError:
            raise DatabaseValidationError(f'{obj} is not in the database.')
        result = self.refs.pop(index)
        self._unset_database(result)
        return result

    def delete_enum(self, obj: Enum) -> Enum:
        try:
            index = self.enums.index(obj)
        except ValueError:
            raise DatabaseValidationError(f'{obj} is not in the database.')
        result = self.enums.pop(index)
        self._unset_database(result)
        return result

    def delete_table_group(self, obj: TableGroup) -> TableGroup:
        try:
            index = self.table_groups.index(obj)
        except ValueError:
            raise DatabaseValidationError(f'{obj} is not in the database.')
        result = self.table_groups.pop(index)
        self._unset_database(result)
        return result

    def delete_project(self) -> Project:
        if self.project is None:
            raise DatabaseValidationError(f'Project is not set.')
        result = self.project
        self.project = None
        self._unset_database(result)
        return result

    @property
    def sql(self):
        '''Returs SQL of the parsed results'''
        return self.sql_renderer.render_db(self)

    @property
    def dbml(self):
        '''Generates DBML code out of parsed results'''
        return self.dbml_renderer.render_db(self)



class TableNotFoundError(Exception):
    pass


class ColumnNotFoundError(Exception):
    pass


class IndexNotFoundError(Exception):
    pass


class AttributeMissingError(Exception):
    pass


class DuplicateReferenceError(Exception):
    pass


class UnknownDatabaseError(Exception):
    pass


class DBMLError(Exception):
    pass


class DatabaseValidationError(Exception):
    pass


class ValidationError(Exception):
    pass



from textwrap import indent

from pydbml.classes import TableGroup
from pydbml.renderer.dbml.default.renderer import DefaultDBMLRenderer
from pydbml.renderer.dbml.default.table import get_full_name_for_dbml
from pydbml.renderer.dbml.default.utils import comment_to_dbml
from pydbml.tools import doublequote_string


@DefaultDBMLRenderer.renderer_for(TableGroup)
def render_table_group(model: TableGroup) -> str:
    result = comment_to_dbml(model.comment) if model.comment else ''
    quoted_name = doublequote_string(model.name)
    result += f'TableGroup {quoted_name}'
    if model.color:
        result += f' [color: {model.color}]'
    result += ' {\n'
    for i in model.items:
        result += f'    {get_full_name_for_dbml(i)}\n'
    if model.note:
        result += indent(model.note.dbml, '    ') + '\n'
    result += '}'
    return result



from .base import SQLObject, DBMLObject


class Expression(SQLObject, DBMLObject):
    def __init__(self, text: str):
        self.text = text

    def __str__(self) -> str:
        '''
        >>> print(Expression('sum(amount)'))
        sum(amount)
        '''

        return self.text

    def __repr__(self) -> str:
        '''
        >>> Expression('sum(amount)')
        Expression('sum(amount)')
        '''

        return f'Expression({repr(self.text)})'



from typing import Type, Callable, Dict, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from pydbml.database import Database


def unsupported_renderer(model) -> str:
    return ''


class BaseRenderer:
    _unsupported_renderer = unsupported_renderer

    @property
    def model_renderers(cls) -> Dict[Type, Callable]:
        """A class attribute dictionary to store the model renderers."""
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def render(cls, model) -> str:
        """
        Render the model to a string. If the model is not supported, fall back to
        `self._unsupported_renderer` that by default returns an empty string.
        """

        return cls.model_renderers.get(type(model), cls._unsupported_renderer)(model)  # type: ignore

    @classmethod
    def renderer_for(cls, model_cls: Type) -> Callable:
        """A decorator to register a renderer for a model class."""
        def decorator(func) -> Callable:
            cls.model_renderers[model_cls] = func  # type: ignore
            return func
        return decorator

    @classmethod
    def render_db(cls, db: 'Database') -> str:
        raise NotImplementedError  # pragma: no cover



from pydbml.classes import Expression
from pydbml.renderer.dbml.default.renderer import DefaultDBMLRenderer


@DefaultDBMLRenderer.renderer_for(Expression)
def render_expression(model: Expression) -> str:
    return f'`{model.text}`'



import pyparsing as pp

from .common import _, end, _c
from .generic import string_literal, name
from ..parser.blueprints import StickyNoteBlueprint

sticky_note = _c + pp.CaselessLiteral('note') + _ + (name('name') + _ - '{' + _ - string_literal('text') + _ - '}') + end


def parse_sticky_note(s, loc, tok):
    '''
    Note single_line_note {
      'This is a single line note'
    }
    '''
    init_dict = {'name': tok['name'], 'text': tok['text']}

    return StickyNoteBlueprint(**init_dict)


sticky_note.set_parse_action(parse_sticky_note)



from typing import TYPE_CHECKING, List

from pydbml.renderer.base import BaseRenderer
from pydbml._classes.base import DBMLObject

if TYPE_CHECKING:  # pragma: no cover
    from pydbml.database import Database


class DefaultDBMLRenderer(BaseRenderer):
    model_renderers = {}

    @classmethod
    def render_db(cls, db: 'Database') -> str:
        items: List[DBMLObject] = [db.project] if db.project else []
        refs = (ref for ref in db.refs if not ref.inline)
        items.extend((*db.enums, *db.tables, *refs, *db.table_groups, *db.sticky_notes))

        return '\n\n'.join(cls.render(i) for i in items)



from typing import List, Dict, Union

from pydbml.classes import Enum, Reference, Table
from pydbml.constants import MANY_TO_ONE, ONE_TO_MANY
from pydbml.tools import comment


def comment_to_sql(val: str) -> str:
    return comment(val, '--')


def reorder_tables_for_sql(tables: List['Table'], refs: List['Reference']) -> List['Table']:
    """
    Attempt to reorder the tables, so that they are defined in SQL before they are referenced by
    inline foreign keys.

    Won't aid the rare cases of cross-references and many-to-many relations.
    """

    references: Dict[str, int] = {}
    for ref in refs:
        if ref.inline:
            if ref.type == MANY_TO_ONE and ref.table1 is not None:
                table_name = ref.table1.name
            elif ref.type == ONE_TO_MANY and ref.table2 is not None:
                table_name = ref.table2.name
            else:  # pragma: no cover
                continue
            references[table_name] = references.get(table_name, 0) + 1
    return sorted(tables, key=lambda t: references.get(t.name, 0), reverse=True)


def get_full_name_for_sql(model: Union[Table, Enum]) -> str:
    if model.schema == 'public':
        return f'"{model.name}"'
    else:
        return f'"{model.schema}"."{model.name}"'



from .._classes.column import Column
from .._classes.enum import Enum
from .._classes.enum import EnumItem
from .._classes.expression import Expression
from .._classes.index import Index
from .._classes.note import Note
from .._classes.project import Project
from .._classes.reference import Reference
from .._classes.sticky_note import StickyNote
from .._classes.table import Table
from .._classes.table_group import TableGroup

__all__ = [
    "Column",
    "Enum",
    "EnumItem",
    "Expression",
    "Index",
    "Note",
    "Project",
    "Reference",
    "StickyNote",
    "Table",
    "TableGroup",
]



from typing import TYPE_CHECKING

from pydbml.renderer.sql.default.utils import reorder_tables_for_sql
from pydbml.renderer.base import BaseRenderer


if TYPE_CHECKING:  # pragma: no cover
    from pydbml.database import Database


class DefaultSQLRenderer(BaseRenderer):
    model_renderers = {}

    @classmethod
    def render(cls, model) -> str:
        model.check_attributes_for_sql()
        return super().render(model)

    @classmethod
    def render_db(cls, db: 'Database') -> str:
        refs = (ref for ref in db.refs if not ref.inline)
        tables = reorder_tables_for_sql(db.tables, db.refs)
        components = (cls.render(i) for i in (*db.enums, *tables, *refs))
        return '\n\n'.join(components)



import pyparsing as pp

from .common import _
from .common import _c
from .common import c
from .common import n
from .generic import name
from pydbml.parser.blueprints import ReferenceBlueprint

pp.ParserElement.set_default_whitespace_chars(' \t\r')

relation = pp.oneOf("> - < <>")

col_name = (
    (
        name('schema') + '.' + name('table') + '.' - name('field')
    ) | (
        name('table') + '.' + name('field')
    )
)

ref_inline = pp.Literal("ref:") - relation('type') - col_name


def parse_inline_relation(s, loc, tok):
    '''
    ref: < table.column
    or
    ref: < schema1.table.column
    '''
    result = {
        'type': tok['type'],
        'inline': True,
        'table2': tok['table'],
        'col2': tok['field']
    }
    if 'schema' in tok:
        result['schema2'] = tok['schema']
    return ReferenceBlueprint(**result)


ref_inline.set_parse_action(parse_inline_relation)

on_option = (
    pp.CaselessLiteral('no action')
    | pp.CaselessLiteral('restrict')
    | pp.CaselessLiteral('cascade')
    | pp.CaselessLiteral('set null')
    | pp.CaselessLiteral('set default')
)
update = pp.CaselessLiteral("update:").suppress() + _ + on_option
delete = pp.CaselessLiteral("delete:").suppress() + _ + on_option

ref_setting = _ + (update('update') | delete('delete')) + _

ref_settings = (
    '['
    + ref_setting
    + (
        ','
        + ref_setting
    )[...]
    + ']' + c
)


def parse_ref_settings(s, loc, tok):
    '''
    [delete: cascade]
    '''
    result = {}
    if 'update' in tok:
        result['on_update'] = tok['update'][0]
    if 'delete' in tok:
        result['on_delete'] = tok['delete'][0]
    if 'comment' in tok:
        result['comment'] = tok['comment'][0]
    return result


ref_settings.set_parse_action(parse_ref_settings)

composite_name = (
    '(' + pp.White()[...]
    - name + pp.White()[...]
    + (
        pp.White()[...] + ","
        + pp.White()[...] + name
        + pp.White()[...]
    )[...]
    + ')'
)
name_or_composite = name | pp.Combine(composite_name)

ref_cols = (
    (
        name('schema')
        + pp.Suppress('.') + name('table')
        + pp.Suppress('.') + name_or_composite('field')
    ) | (
        name('table')
        + pp.Suppress('.') + name_or_composite('field')
    )
)


def parse_ref_cols(s, loc, tok):
    '''
    table1.col1
    or
    schema1.table1.col1
    or
    schema1.table1.(col1, col2)
    '''
    result = {
        'table': tok['table'],
        'field': tok['field'],
    }
    if 'schema' in tok:
        result['schema'] = tok['schema']
    return result


ref_cols.set_parse_action(parse_ref_cols)

ref_body = (
    ref_cols('col1')
    - relation('type')
    - ref_cols('col2') + c
    + ref_settings('settings')[0, 1]
)
# ref_body = (
#     table_name('table1')
#     - '.'
#     - name_or_composite('field1')
#     - relation('type')
#     - table_name('table2')
#     - '.'
#     - name_or_composite('field2') + c
#     + ref_settings('settings')[0, 1]
# )


ref_short = _c + pp.CaselessLiteral('ref') + name('name')[0, 1] + ':' - ref_body
ref_long = _c + (
    pp.CaselessLiteral('ref') + _
    + name('name')[0, 1] + _
    + '{' + _
    - ref_body + _
    - '}'
)


def parse_ref(s, loc, tok):
    '''
    ref name: table1.col1 > table2.col2
    or
    ref name {
        table1.col1 < table2.col2
    }
    '''
    init_dict = {
        'type': tok['type'],
        'inline': False,
        'table1': tok['col1']['table'],
        'col1': tok['col1']['field'],
        'table2': tok['col2']['table'],
        'col2': tok['col2']['field'],
    }

    if 'schema' in tok['col1']:
        init_dict['schema1'] = tok['col1']['schema']
    if 'schema' in tok['col2']:
        init_dict['schema2'] = tok['col2']['schema']
    if 'name' in tok:
        init_dict['name'] = tok['name']
    if 'settings' in tok:
        init_dict.update(tok['settings'])

    # comments after settings have priority
    if 'comment' in tok:
        init_dict['comment'] = tok['comment'][0]
    if 'comment' not in init_dict and 'comment_before' in tok:
        comment = '\n'.join(c[0] for c in tok['comment_before'])
        init_dict['comment'] = comment

    ref = ReferenceBlueprint(**init_dict)
    return ref


ref_short.set_parse_action(parse_ref)
ref_long.set_parse_action(parse_ref)

ref = ref_short | ref_long + (n | pp.StringEnd())



from textwrap import indent

from pydbml._classes.enum import EnumItem
from pydbml.classes import Enum
from pydbml.renderer.sql.default.renderer import DefaultSQLRenderer
from pydbml.renderer.sql.default.utils import comment_to_sql, get_full_name_for_sql


@DefaultSQLRenderer.renderer_for(Enum)
def render_enum(model: Enum) -> str:
    '''
    Returns SQL for enum type:

    CREATE TYPE "job_status" AS ENUM (
      'created',
      'running',
      'done',
      'failure',
    );
    '''

    result = comment_to_sql(model.comment) if model.comment else ''
    result += f'CREATE TYPE {get_full_name_for_sql(model)} AS ENUM (\n'
    result += '\n'.join(f'{indent(DefaultSQLRenderer.render(i), "  ")}' for i in model.items)
    result += '\n);'
    return result


@DefaultSQLRenderer.renderer_for(EnumItem)
def render_enum_item(model: EnumItem) -> str:
    result = comment_to_sql(model.comment) if model.comment else ''
    result += f"'{model.name}',"
    return result



import pyparsing as pp

from .common import _
from .common import _c
from .common import n
from .common import note
from .common import note_object
from .generic import name
from .generic import string_literal
from pydbml.parser.blueprints import NoteBlueprint
from pydbml.parser.blueprints import ProjectBlueprint

pp.ParserElement.set_default_whitespace_chars(' \t\r')

project_field = pp.Group(name + _ + pp.Suppress(':') + _ - string_literal)

project_element = _ + (note | note_object | project_field) + _

project_body = project_element[...]

project = _c + (
    pp.CaselessLiteral('project') + _
    - name('name') + _
    + '{' + _
    - project_body('items') + _
    - '}'
) + (n | pp.StringEnd())


def parse_project(s, loc, tok):
    '''
    Project project_name {
      database_type: 'PostgreSQL'
      Note: 'Description of the project'
    }
    '''
    init_dict = {'name': tok['name']}
    items = {}
    for item in tok.get('items', []):
        if isinstance(item, NoteBlueprint):
            init_dict['note'] = item
        else:
            k, v = item
            items[k] = v
    if items:
        init_dict['items'] = items
    if 'comment_before' in tok:
        comment = '\n'.join(c[0] for c in tok['comment_before'])
        init_dict['comment'] = comment
    return ProjectBlueprint(**init_dict)


project.set_parse_action(parse_project)



import os

from pathlib import Path
from unittest import TestCase
from unittest.mock import Mock

from pydbml.classes import Column
from pydbml.classes import Enum
from pydbml.classes import EnumItem
from pydbml.classes import Project
from pydbml.classes import Reference
from pydbml.classes import Table
from pydbml.classes import TableGroup
from pydbml.database import Database
from pydbml.exceptions import DatabaseValidationError
from pydbml.constants import ONE_TO_MANY, MANY_TO_ONE, MANY_TO_MANY
from pydbml.renderer.sql.default.utils import reorder_tables_for_sql

TEST_DATA_PATH = Path(os.path.abspath(__file__)).parent / 'test_data'


class TestDatabase(TestCase):
    def test_add_table(self) -> None:
        c = Column('test', 'varchar', True)
        t = Table('test_table')
        t.add_column(c)
        database = Database()
        res = database.add_table(t)
        self.assertEqual(t.database, database)
        self.assertIs(res, t)
        self.assertIn(t, database.tables)

    def test_add_table_alias(self) -> None:
        c = Column('test', 'varchar', True)
        t = Table('test_table', alias='myalias')
        t.add_column(c)
        database = Database()
        database.add_table(t)
        self.assertIsInstance(t.alias, str)
        self.assertIs(database[t.alias], t)

    def test_add_table_alias_bad(self) -> None:
        c = Column('test', 'varchar', True)
        t = Table('test', alias='myalias')
        t.add_column(c)
        database = Database()
        database.add_table(t)
        t2 = Table('test_table', alias='myalias')
        with self.assertRaises(DatabaseValidationError):
            database.add_table(t2)
        self.assertIsNone(t2.database)

    def test_add_table_bad(self) -> None:
        c = Column('test', 'varchar', True)
        t = Table('test_table')
        t.add_column(c)
        database = Database()
        database.add_table(t)
        with self.assertRaises(DatabaseValidationError):
            database.add_table(t)
        t2 = Table('test_table')
        with self.assertRaises(DatabaseValidationError):
            database.add_table(t2)

    def test_delete_table(self) -> None:
        c = Column('test', 'varchar', True)
        t = Table('test_table', alias='myalias')
        t.add_column(c)
        database = Database()
        database.add_table(t)
        res = database.delete_table(t)
        self.assertIsNone(t.database, database)
        self.assertIs(res, t)
        self.assertNotIn(t, database.tables)
        self.assertNotIn('test_table', database.table_dict)
        self.assertNotIn('myalias', database.table_dict)

    def test_delete_missing_table(self) -> None:
        t = Table('test_table')
        database = Database()
        with self.assertRaises(DatabaseValidationError):
            database.delete_table(t)
        self.assertIsNone(t.database, database)

    def test_add_reference(self) -> None:
        c = Column('test', 'varchar', True)
        t = Table('test_table')
        t.add_column(c)
        database = Database()
        database.add_table(t)
        c2 = Column('test2', 'integer')
        t2 = Table('test_table2')
        t2.add_column(c2)
        database.add_table(t2)
        ref = Reference('>', c, c2)
        res = database.add_reference(ref)
        self.assertEqual(ref.database, database)
        self.assertIs(res, ref)
        self.assertIn(ref, database.refs)

    def test_add_reference_bad(self) -> None:
        c = Column('test', 'varchar', True)
        t = Table('test_table')
        t.add_column(c)
        database = Database()
        database.add_table(t)
        c2 = Column('test2', 'integer')
        t2 = Table('test_table2')
        t2.add_column(c2)
        database.add_table(t2)
        ref = Reference('>', c, c2)
        database.add_reference(ref)
        with self.assertRaises(DatabaseValidationError):
            database.add_reference(ref)

        c3 = Column('test', 'varchar', True)
        t3 = Table('test_table')
        t3.add_column(c3)
        database3 = Database()
        database3.add_table(t3)
        c32 = Column('test2', 'integer')
        t32 = Table('test_table2')
        t32.add_column(c32)
        database3.add_table(t32)
        ref3 = Reference('>', c3, c32)
        with self.assertRaises(DatabaseValidationError):
            database.add_reference(ref3)

    def test_delete_reference(self) -> None:
        c = Column('test', 'varchar', True)
        t = Table('test_table')
        t.add_column(c)
        database = Database()
        database.add_table(t)
        c2 = Column('test2', 'integer')
        t2 = Table('test_table2')
        t2.add_column(c2)
        database.add_table(t2)
        ref = Reference('>', c, c2)
        res = database.add_reference(ref)
        res = database.delete_reference(ref)
        self.assertIsNone(ref.database, database)
        self.assertIs(res, ref)
        self.assertNotIn(ref, database.refs)

    def test_delete_missing_reference(self) -> None:
        c = Column('test', 'varchar', True)
        t = Table('test_table')
        t.add_column(c)
        database = Database()
        database.add_table(t)
        c2 = Column('test2', 'integer')
        t2 = Table('test_table2')
        t2.add_column(c2)
        database.add_table(t2)
        ref = Reference('>', c, c2)
        with self.assertRaises(DatabaseValidationError):
            database.delete_reference(ref)
        self.assertIsNone(ref.database)

    def test_add_enum(self) -> None:
        e = Enum('myenum', [EnumItem('a'), EnumItem('b')])
        database = Database()
        res = database.add_enum(e)
        self.assertEqual(e.database, database)
        self.assertIs(res, e)
        self.assertIn(e, database.enums)

    def test_add_enum_bad(self) -> None:
        e = Enum('myenum', [EnumItem('a'), EnumItem('b')])
        database = Database()
        database.add_enum(e)
        with self.assertRaises(DatabaseValidationError):
            database.add_enum(e)
        e2 = Enum('myenum', [EnumItem('a2'), EnumItem('b2')])
        with self.assertRaises(DatabaseValidationError):
            database.add_enum(e2)

    def test_delete_enum(self) -> None:
        e = Enum('myenum', [EnumItem('a'), EnumItem('b')])
        database = Database()
        database.add_enum(e)
        res = database.delete_enum(e)
        self.assertIsNone(e.database)
        self.assertIs(res, e)
        self.assertNotIn(e, database.enums)

    def test_delete_missing_enum(self) -> None:
        e = Enum('myenum', [EnumItem('a'), EnumItem('b')])
        database = Database()
        with self.assertRaises(DatabaseValidationError):
            database.delete_enum(e)
        self.assertIsNone(e.database)

    def test_add_table_group(self) -> None:
        t1 = Table('table1')
        t2 = Table('table2')
        tg = TableGroup('mytablegroup', [t1, t2])
        database = Database()
        res = database.add_table_group(tg)
        self.assertEqual(tg.database, database)
        self.assertIs(res, tg)
        self.assertIn(tg, database.table_groups)

    def test_add_table_group_bad(self) -> None:
        t1 = Table('table1')
        t2 = Table('table2')
        tg = TableGroup('mytablegroup', [t1, t2])
        database = Database()
        database.add_table_group(tg)
        with self.assertRaises(DatabaseValidationError):
            database.add_table_group(tg)
        tg2 = TableGroup('mytablegroup', [t2])
        with self.assertRaises(DatabaseValidationError):
            database.add_table_group(tg2)

    def test_delete_table_group(self) -> None:
        t1 = Table('table1')
        t2 = Table('table2')
        tg = TableGroup('mytablegroup', [t1, t2])
        database = Database()
        database.add_table_group(tg)
        res = database.delete_table_group(tg)
        self.assertIsNone(tg.database)
        self.assertIs(res, tg)
        self.assertNotIn(tg, database.table_groups)

    def test_delete_missing_table_group(self) -> None:
        t1 = Table('table1')
        t2 = Table('table2')
        tg = TableGroup('mytablegroup', [t1, t2])
        database = Database()
        with self.assertRaises(DatabaseValidationError):
            database.delete_table_group(tg)
        self.assertIsNone(tg.database)

    def test_add_project(self) -> None:
        p = Project('myproject')
        database = Database()
        res = database.add_project(p)
        self.assertEqual(p.database, database)
        self.assertIs(res, p)
        self.assertIs(database.project, p)

    def test_add_another_project(self) -> None:
        p = Project('myproject')
        database = Database()
        database.add_project(p)
        p2 = Project('anotherproject')
        res = database.add_project(p2)
        self.assertEqual(p2.database, database)
        self.assertIs(res, p2)
        self.assertIs(database.project, p2)
        self.assertIsNone(p.database)

    def test_delete_project(self) -> None:
        p = Project('myproject')
        database = Database()
        database.add_project(p)
        res = database.delete_project()
        self.assertIsNone(p.database, database)
        self.assertIs(res, p)
        self.assertIsNone(database.project)

    def test_delete_missing_project(self) -> None:
        database = Database()
        with self.assertRaises(DatabaseValidationError):
            database.delete_project()

    def test_geititem(self) -> None:
        t1 = Table('table1')
        t2 = Table('table2', schema='myschema')
        database = Database()
        database.add_table(t1)
        database.add_table(t2)
        self.assertIs(database['public.table1'], t1)
        self.assertIs(database['myschema.table2'], t2)
        self.assertIs(database[0], t1)
        self.assertIs(database[1], t2)
        with self.assertRaises(TypeError):
            database[None]
        with self.assertRaises(IndexError):
            database[2]
        with self.assertRaises(KeyError):
            database['wrong']

    def test_iter(self) -> None:
        t1 = Table('table1')
        t2 = Table('table2')
        database = Database()
        database.add_table(t1)
        database.add_table(t2)
        self.assertEqual(list(iter(database)), [t1, t2])

    def test_add(self) -> None:
        t1 = Table('table1')
        t2 = Table('table2')
        tg = TableGroup('mytablegroup', [t1, t2])
        e = Enum('myenum', [EnumItem('a'), EnumItem('b')])
        database = Database()
        database.add(t1)
        database.add(t2)
        database.add(e)
        database.add(tg)
        self.assertIs(t1.database, database)
        self.assertIs(t2.database, database)
        self.assertIs(e.database, database)
        self.assertIs(tg.database, database)
        self.assertIn(t1, database.tables)
        self.assertIn(t2, database.tables)
        self.assertIn(tg, database.table_groups)
        self.assertIn(e, database.enums)

    def test_add_bad(self) -> None:
        class Test:
            pass
        t = Test()
        database = Database()
        with self.assertRaises(DatabaseValidationError):
            database.add(t)
        with self.assertRaises(AttributeError):
            t.database

    def test_delete(self) -> None:
        t1 = Table('table1')
        c1 = Column('col1', 'int')
        t1.add_column(c1)
        t2 = Table('table2')
        c2 = Column('col2', 'int')
        t2.add_column(c2)
        ref = Reference('>', [c1], [c2])
        tg = TableGroup('mytablegroup', [t1, t2])
        e = Enum('myenum', [EnumItem('a'), EnumItem('b')])
        p = Project('myproject')
        database = Database()
        database.add(t1)
        database.add(t2)
        database.add(e)
        database.add(tg)
        database.add(ref)
        database.add(p)

        database.delete(t1)
        database.delete(t2)
        database.delete(e)
        database.delete(tg)
        database.delete(ref)
        database.delete(p)
        self.assertIsNone(t1.database)
        self.assertIsNone(t2.database)
        self.assertIsNone(e.database)
        self.assertIsNone(tg.database)
        self.assertIsNone(ref.database)
        self.assertIsNone(p.database)
        self.assertIsNone(database.project)
        self.assertNotIn(t1, database.tables)
        self.assertNotIn(t2, database.tables)
        self.assertNotIn(tg, database.table_groups)
        self.assertNotIn(e, database.enums)
        self.assertNotIn(ref, database.refs)

    def test_delete_bad(self) -> None:
        class Test:
            pass
        t = Test()
        database = Database()
        with self.assertRaises(DatabaseValidationError):
            database.delete(t)
        with self.assertRaises(AttributeError):
            t.database


def test_repr() -> None:
    assert repr(Database()) == "<Database>"



from pydbml.classes import Expression
from pydbml.renderer.sql.default.renderer import DefaultSQLRenderer


@DefaultSQLRenderer.renderer_for(Expression)
def render_expression(model: Expression) -> str:
    return f'({model.text})'



from .column import render_column
from .enum import render_enum, render_enum_item
from .expression import render_expression
from .index import render_index
from .note import render_note
from .project import render_project
from .reference import render_reference
from .renderer import DefaultDBMLRenderer
from .sticky_note import render_sticky_note
from .table import render_table
from .table_group import render_table_group



from __future__ import annotations

from io import TextIOWrapper
from pathlib import Path
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import pyparsing as pp

from pydbml.classes import Table
from pydbml.database import Database
from pydbml.definitions.common import comment
from pydbml.definitions.enum import enum
from pydbml.definitions.project import project
from pydbml.definitions.reference import ref
from pydbml.definitions.sticky_note import sticky_note
from pydbml.definitions.table import table, table_with_properties
from pydbml.definitions.table_group import table_group
from pydbml.exceptions import TableNotFoundError
from pydbml.renderer.base import BaseRenderer
from pydbml.renderer.dbml.default import DefaultDBMLRenderer
from pydbml.renderer.sql.default import DefaultSQLRenderer
from pydbml.tools import remove_bom
from .blueprints import EnumBlueprint, StickyNoteBlueprint
from .blueprints import ProjectBlueprint
from .blueprints import ReferenceBlueprint
from .blueprints import TableBlueprint
from .blueprints import TableGroupBlueprint

pp.ParserElement.set_default_whitespace_chars(" \t\r")


class PyDBML:
    """
    PyDBML parser factory. If properly initiated, returns parsed Database.

    Usage option 1:

    >>> with open('test_schema.dbml') as f:
    ...     p = PyDBML(f)
    ...     # or
    ...     p = PyDBML(f.read())

    Usage option 2:
    >>> p = PyDBML.parse_file('test_schema.dbml')
    >>> # or
    >>> from pathlib import Path
    >>> p = PyDBML(Path('test_schema.dbml'))
    """

    def __new__(
        cls,
        source_: Optional[Union[str, Path, TextIOWrapper]] = None,
        allow_properties: bool = False,
        sql_renderer: Type[BaseRenderer] = DefaultSQLRenderer,
        dbml_renderer: Type[BaseRenderer] = DefaultDBMLRenderer,
    ):
        if source_ is not None:
            if isinstance(source_, str):
                source = source_
            elif isinstance(source_, Path):
                with open(source_, encoding="utf8") as f:
                    source = f.read()
            elif isinstance(source_, TextIOWrapper):
                source = source_.read()
            else:
                raise TypeError("Source must be str, path or file stream")

            source = remove_bom(source)
            return cls.parse(
                source,
                allow_properties=allow_properties,
                sql_renderer=sql_renderer,
                dbml_renderer=dbml_renderer,
            )
        else:
            return super().__new__(cls)

    def __repr__(self):
        return "<PyDBML>"

    @staticmethod
    def parse(
        text: str,
        allow_properties: bool = False,
        sql_renderer: Type[BaseRenderer] = DefaultSQLRenderer,
        dbml_renderer: Type[BaseRenderer] = DefaultDBMLRenderer,
    ) -> Database:
        text = remove_bom(text)
        parser = PyDBMLParser(
            text,
            allow_properties=allow_properties,
            sql_renderer=sql_renderer,
            dbml_renderer=dbml_renderer,
        )
        return parser.parse()

    @staticmethod
    def parse_file(file: Union[str, Path, TextIOWrapper]) -> Database:
        if isinstance(file, TextIOWrapper):
            source = file.read()
        else:
            with open(file, encoding="utf8") as f:
                source = f.read()
        source = remove_bom(source)
        parser = PyDBMLParser(source)
        return parser.parse()


class PyDBMLParser:
    def __init__(
        self,
        source: str,
        allow_properties: bool = False,
        sql_renderer: Type[BaseRenderer] = DefaultSQLRenderer,
        dbml_renderer: Type[BaseRenderer] = DefaultDBMLRenderer,
    ):
        self.database = None

        self.ref_blueprints: List[ReferenceBlueprint] = []
        self.table_groups: List[TableGroupBlueprint] = []
        self.source = source
        self.tables: List[TableBlueprint] = []
        self.refs: List[ReferenceBlueprint] = []
        self.enums: List[EnumBlueprint] = []
        self.project: Optional[ProjectBlueprint] = None
        self.sticky_notes: List[StickyNoteBlueprint] = []
        self._allow_properties = allow_properties
        self._sql_renderer = sql_renderer
        self._dbml_renderer = dbml_renderer

    def parse(self):
        self._set_syntax()
        self._syntax.parse_string(self.source, parseAll=True)
        self.build_database()
        return self.database

    def __repr__(self):
        return "<PyDBMLParser>"

    def _set_syntax(self):
        table_expr = (
            table_with_properties.copy() if self._allow_properties else table.copy()
        )
        ref_expr = ref.copy()
        enum_expr = enum.copy()
        table_group_expr = table_group.copy()
        project_expr = project.copy()
        note_expr = sticky_note.copy()

        table_expr.addParseAction(self.parse_blueprint)
        ref_expr.addParseAction(self.parse_blueprint)
        enum_expr.addParseAction(self.parse_blueprint)
        table_group_expr.addParseAction(self.parse_blueprint)
        project_expr.addParseAction(self.parse_blueprint)
        note_expr.addParseAction(self.parse_blueprint)

        expr = (
            table_expr
            | ref_expr
            | enum_expr
            | table_group_expr
            | project_expr
            | note_expr
        )
        self._syntax = expr[...] + ("\n" | comment)[...] + pp.StringEnd()

    def parse_blueprint(self, s, loc, tok):
        blueprint = tok[0]
        if isinstance(blueprint, TableBlueprint):
            self.tables.append(blueprint)
            ref_bps = blueprint.get_reference_blueprints()
            col_bps = blueprint.columns or []
            index_bps = blueprint.indexes or []
            for ref_bp in ref_bps:
                self.refs.append(ref_bp)
                ref_bp.parser = self
            for col_bp in col_bps:
                col_bp.parser = self
                if col_bp.note:
                    col_bp.note.parser = self
            for index_bp in index_bps:
                index_bp.parser = self
                if index_bp.note:
                    index_bp.note.parser = self
            if blueprint.note:
                blueprint.note.parser = self
        elif isinstance(blueprint, ReferenceBlueprint):
            self.refs.append(blueprint)
        elif isinstance(blueprint, EnumBlueprint):
            self.enums.append(blueprint)
            for enum_item in blueprint.items:
                if enum_item.note:
                    enum_item.note.parser = self
        elif isinstance(blueprint, TableGroupBlueprint):
            self.table_groups.append(blueprint)
        elif isinstance(blueprint, ProjectBlueprint):
            self.project = blueprint
            if blueprint.note:
                blueprint.note.parser = self
        elif isinstance(blueprint, StickyNoteBlueprint):
            self.sticky_notes.append(blueprint)
        else:
            raise RuntimeError(f"type unknown: {blueprint}")
        blueprint.parser = self

    def locate_table(self, schema: str, name: str) -> "Table":
        if not self.database:
            raise RuntimeError("Database is not ready")
        # first by alias
        result = self.database.table_dict.get(name)
        if result is None:
            full_name = f"{schema}.{name}"
            result = self.database.table_dict.get(full_name)
        if result is None:
            raise TableNotFoundError(f"Table {full_name} not present in the database")
        return result

    def build_database(self):
        self.database = Database(
            allow_properties=self._allow_properties,
            sql_renderer=self._sql_renderer,
            dbml_renderer=self._dbml_renderer,
        )
        for enum_bp in self.enums:
            self.database.add(enum_bp.build())
        for table_bp in self.tables:
            self.database.add(table_bp.build())
            self.ref_blueprints.extend(table_bp.get_reference_blueprints())
        for table_group_bp in self.table_groups:
            self.database.add(table_group_bp.build())
        for note_bp in self.sticky_notes:
            self.database.add(note_bp.build())
        if self.project:
            self.database.add(self.project.build())
        for ref_bp in self.refs:
            self.database.add(ref_bp.build())



import pyparsing as pp

from .common import _
from .common import _c
from .common import c
from .common import note
from .common import pk
from .common import unique
from .generic import expression_literal
from .generic import name
from .generic import string_literal
from pydbml.parser.blueprints import ExpressionBlueprint
from pydbml.parser.blueprints import IndexBlueprint

pp.ParserElement.set_default_whitespace_chars(' \t\r')

index_type = pp.CaselessLiteral("type:").suppress() + _ - (
    pp.CaselessLiteral("brin")('type') |
    pp.CaselessLiteral("btree")('type') |
    pp.CaselessLiteral("gin")('type') |
    pp.CaselessLiteral("gist")('type') |
    pp.CaselessLiteral("hash")('type') |
    pp.CaselessLiteral("spgist")('type')
)
index_setting = _ + (
    unique('unique')
    | index_type
    | pp.CaselessLiteral("name:") + _ - string_literal('name')
    | note('note')
    | pk('pk')
) + _
index_settings = (
    '[' + index_setting + (',' - index_setting)[...] - ']' + c
)


def parse_index_settings(s, lok, tok):
    '''
    [type: btree, name: 'name', unique, note: 'note']
    '''
    result = {}
    if 'unique' in tok:
        result['unique'] = True
    if 'name' in tok:
        result['name'] = tok['name']
    if 'pk' in tok:
        result['pk'] = True
    if 'type' in tok:
        result['type'] = tok['type']
    if 'note' in tok:
        result['note'] = tok['note']
    if 'comment' in tok:
        result['comment'] = tok['comment'][0]
    return result


index_settings.set_parse_action(parse_index_settings)

subject = name | expression_literal
composite_index_syntax = (
    pp.Suppress('(')
    + subject + (
        pp.Suppress(',')
        + subject
    )[...]
    + pp.Suppress(')')
)('subject') + c + index_settings('settings')[0, 1]

single_index_syntax = subject('subject') + c + index_settings('settings')[0, 1]
index = _c + (single_index_syntax ^ composite_index_syntax) + c

indexes = (
    pp.CaselessLiteral('indexes').suppress() + _
    - pp.Suppress('{')
    - index[1, ...] + _
    + pp.Suppress('}')
)


def parse_index(s, lok, tok):
    '''
        (id, country) [pk] // composite primary key
        or
        "created_at"
        or
        booking_date [
          name: 'name',
          unique
        ]
    '''
    init_dict = {}
    if isinstance(tok['subject'], (str, ExpressionBlueprint)):
        subjects = [tok['subject']]
    else:
        subjects = list(tok['subject'])

    init_dict['subject_names'] = subjects
    settings = tok.get('settings', {})
    init_dict.update(settings)

    # comments after settings have priority
    if 'comment' in tok:
        init_dict['comment'] = tok['comment'][0]
    if 'comment' not in init_dict and 'comment_before' in tok:
        comment = '\n'.join(c[0] for c in tok['comment_before'])
        init_dict['comment'] = comment
    return IndexBlueprint(**init_dict)


index.set_parse_action(parse_index)



from pydbml.classes import Column, Enum, Expression
from pydbml.renderer.sql.default.renderer import DefaultSQLRenderer
from .utils import comment_to_sql
from .enum import get_full_name_for_sql as get_full_name_for_sql_enum


@DefaultSQLRenderer.renderer_for(Column)
def render_column(model: Column) -> str:
    '''
    Returns inline SQL of the column, which should be a part of table definition:

    "id" integer PRIMARY KEY AUTOINCREMENT
    '''

    components = [f'"{model.name}"']
    if isinstance(model.type, Enum):
        components.append(get_full_name_for_sql_enum(model.type))
    else:
        components.append(str(model.type))

    table_has_composite_pk = model.table._has_composite_pk() if model.table else False
    if model.pk and not table_has_composite_pk:  # composite PKs are rendered in table sql
        components.append('PRIMARY KEY')
    if model.autoinc:
        components.append('AUTOINCREMENT')
    if model.unique:
        components.append('UNIQUE')
    if model.not_null:
        components.append('NOT NULL')
    if model.default is not None:
        if isinstance(model.default, Expression):
            default = DefaultSQLRenderer.render(model.default)
        else:
            default = model.default  # type: ignore
        components.append(f'DEFAULT {default}')

    result = comment_to_sql(model.comment) if model.comment else ''
    result += ' '.join(components)
    return result



import os

from pathlib import Path
from unittest import TestCase

from pydbml import PyDBML
from pydbml.exceptions import ColumnNotFoundError
from pydbml.exceptions import TableNotFoundError
from pydbml.parser.parser import PyDBMLParser


TEST_DATA_PATH = Path(os.path.abspath(__file__)).parent / 'test_data'


class TestParser(TestCase):
    def setUp(self):
        self.results = PyDBML.parse_file(TEST_DATA_PATH / 'general.dbml')

    def test_table_refs(self) -> None:
        p = PyDBML.parse_file(TEST_DATA_PATH / 'general.dbml')
        r = p['public.orders'].get_refs()
        self.assertEqual(r[0].col2[0].name, 'order_id')
        self.assertEqual(r[0].col1[0].table.name, 'orders')
        self.assertEqual(r[0].col1[0].name, 'id')
        r = p['public.products'].get_refs()
        self.assertEqual(r[1].col1[0].name, 'merchant_id')
        self.assertEqual(r[1].col2[0].table.name, 'merchants')
        self.assertEqual(r[1].col2[0].name, 'id')
        r = p['public.users'].get_refs()
        self.assertEqual(r[0].col1[0].name, 'country_code')
        self.assertEqual(r[0].col2[0].table.name, 'countries')
        self.assertEqual(r[0].col2[0].name, 'code')

    def test_refs(self) -> None:
        p = PyDBML.parse_file(TEST_DATA_PATH / 'general.dbml')
        r = p.refs
        self.assertEqual(r[0].col1[0].table.name, 'orders')
        self.assertEqual(r[0].col1[0].name, 'id')
        self.assertEqual(r[0].col2[0].table.name, 'order_items')
        self.assertEqual(r[0].col2[0].name, 'order_id')
        self.assertEqual(r[2].col1[0].table.name, 'users')
        self.assertEqual(r[2].col1[0].name, 'country_code')
        self.assertEqual(r[2].col2[0].table.name, 'countries')
        self.assertEqual(r[2].col2[0].name, 'code')
        self.assertEqual(r[4].col1[0].table.name, 'products')
        self.assertEqual(r[4].col1[0].name, 'merchant_id')
        self.assertEqual(r[4].col2[0].table.name, 'merchants')
        self.assertEqual(r[4].col2[0].name, 'id')

    def test_inline_refs_schema(self) -> None:
        # Thanks @jens-koster for this example
        source = '''
Table core.pk_tbl {
    pk_col varchar [pk]
}
Table core.fk_tbl {
    fk_col varchar [ref: > core.pk_tbl.pk_col]
}
'''
        p = PyDBMLParser(source)
        p.parse()
        r = p.refs
        pk_tbl = p.tables[0]
        fk_tble = p.tables[1]
        ref = p.refs[0]
        self.assertEqual(ref.table1, fk_tble.name)
        self.assertEqual(ref.table2, pk_tbl.name)
        self.assertEqual(ref.schema1, fk_tble.schema)
        self.assertEqual(ref.schema2, pk_tbl.schema)


class TestRefs(TestCase):
    def test_reference_aliases(self):
        results = PyDBML.parse_file(TEST_DATA_PATH / 'relationships_aliases.dbml')
        posts, reviews, users = results['public.posts'], results['public.reviews'], results['public.users']
        posts2, reviews2, users2 = results['public.posts2'], results['public.reviews2'], results['public.users2']

        rs = results.refs
        self.assertEqual(rs[0].col1[0].table, users)
        self.assertEqual(rs[0].col2[0].table, posts)
        self.assertEqual(rs[1].col1[0].table, users)
        self.assertEqual(rs[1].col2[0].table, reviews)

        self.assertEqual(rs[2].col1[0].table, posts2)
        self.assertEqual(rs[2].col2[0].table, users2)
        self.assertEqual(rs[3].col1[0].table, reviews2)
        self.assertEqual(rs[3].col2[0].table, users2)

    def test_composite_references(self):
        results = PyDBML.parse_file(TEST_DATA_PATH / 'relationships_composite.dbml')
        self.assertEqual(len(results.tables), 4)
        posts, reviews = results['public.posts'], results['public.reviews']
        posts2, reviews2 = results['public.posts2'], results['public.reviews2']

        rs = results.refs
        self.assertEqual(len(rs), 2)

        self.assertEqual(rs[0].col1[0].table, posts)
        self.assertEqual(rs[0].col1, [posts['id'], posts['tag']])
        self.assertEqual(rs[0].col2[0].table, reviews)
        self.assertEqual(rs[0].col2, [reviews['post_id'], reviews['tag']])

        self.assertEqual(rs[1].col1[0].table, posts2)
        self.assertEqual(rs[1].col1, [posts2['id'], posts2['tag']])
        self.assertEqual(rs[1].col2[0].table, reviews2)
        self.assertEqual(rs[1].col2, [reviews2['post_id'], reviews2['tag']])


class TestDBMLReferenceDef(TestCase):
    def test_dbml_reference_def(self):
        results = PyDBML.parse_file(TEST_DATA_PATH / 'dbml_schema_def.dbml')
        self.assertEqual(len(results.tables), 5)
        self.assertEqual(len(results.table_groups), 1)
        self.assertEqual(len(results.enums), 3)


class TestFaulty(TestCase):
    def test_bad_reference(self) -> None:
        with self.assertRaises(TableNotFoundError):
            PyDBML(TEST_DATA_PATH / 'wrong_inline_ref_table.dbml')
        with self.assertRaises(ColumnNotFoundError):
            PyDBML(TEST_DATA_PATH / 'wrong_inline_ref_column.dbml')

    def test_bad_index(self) -> None:
        with self.assertRaises(ColumnNotFoundError):
            PyDBML(TEST_DATA_PATH / 'wrong_index.dbml')


class TestPyDBMLParser(TestCase):
    def test_edge(self) -> None:
        p = PyDBMLParser('')
        with self.assertRaises(RuntimeError):
            p.locate_table('myschema', 'test')
        with self.assertRaises(RuntimeError):
            p.parse_blueprint(1, 1, [1])


class TestNotesIdempotent(TestCase):
    def test_note_is_idempotent(self):
        dbml_source = """
Table test {
    id integer
    Note {
        '''
        Indented note which is actually a Markdown formatted string:
        
        - List item 1
        - Another list item
        
        ```python
        def test():
              print('Hello world!')
              return 1
        ```
        '''
    }
}
"""
        source_text = \
"""Indented note which is actually a Markdown formatted string:

- List item 1
- Another list item

```python
def test():
      print('Hello world!')
      return 1
```"""
        p = PyDBML(dbml_source)
        note = p.tables[0].note
        self.assertEqual(source_text, note.text)

        p_mod = p
        for _ in range(10):
            p_mod = PyDBML(p_mod.dbml)
            note2 = p_mod.tables[0].note
            self.assertEqual(source_text, note2.text)


def test_repr_pydbml() -> None:
    assert repr(PyDBML()) == "<PyDBML>"


def test_repr_pydbml_parser() -> None:
    assert repr(PyDBMLParser('')) == "<PyDBMLParser>"




from typing import Any
from typing import Tuple

from pydbml.exceptions import AttributeMissingError


class SQLObject:
    '''
    Base class for all SQL objects.
    '''
    required_attributes: Tuple[str, ...] = ()
    dont_compare_fields: Tuple[str, ...] = ()

    def check_attributes_for_sql(self):
        '''
        Check if all attributes, required for rendering SQL are set in the
        instance. If some attribute is missing, raise AttributeMissingError
        '''
        for attr in self.required_attributes:
            if getattr(self, attr) is None:
                raise AttributeMissingError(
                    f'Cannot render SQL. Missing required attribute "{attr}".'
                )
    @property
    def sql(self) -> str:
        if hasattr(self, 'database') and self.database is not None:
            renderer = self.database.sql_renderer
        else:
            from pydbml.renderer.sql.default import DefaultSQLRenderer
            renderer = DefaultSQLRenderer

        return renderer.render(self)

    def __setattr__(self, name: str, value: Any):
        """
        Required for type testing with MyPy.
        """
        super().__setattr__(name, value)

    def __eq__(self, other: object) -> bool:
        """
        Two instances of the same SQLObject subclass are equal if all their
        attributes are equal.
        """

        if not isinstance(other, self.__class__):
            return False
        # not comparing those because they are circular references

        self_dict = dict(self.__dict__)
        other_dict = dict(other.__dict__)

        for field in self.dont_compare_fields:
            self_dict.pop(field, None)
            other_dict.pop(field, None)

        return self_dict == other_dict


class DBMLObject:
    '''Base class for all DBML objects.'''
    @property
    def dbml(self) -> str:
        if hasattr(self, 'database') and self.database is not None:
            renderer = self.database.dbml_renderer
        else:
            from pydbml.renderer.dbml.default import DefaultDBMLRenderer
            renderer = DefaultDBMLRenderer

        return renderer.render(self)



import re
from typing import TYPE_CHECKING

from pydbml.tools import comment

if TYPE_CHECKING:  # pragma: no cover
    from pydbml.classes import Note


def prepare_text_for_dbml(text: str) -> str:
    '''Escape single quotes'''
    pattern = re.compile(r"('''|')")
    return pattern.sub(r'\\\1', text)


def quote_string(text: str) -> str:
    if '\n' in text:
        return f"'''\n{prepare_text_for_dbml(text)}'''"
    else:
        return f"'{prepare_text_for_dbml(text)}'"


def note_option_to_dbml(note: 'Note') -> str:
    if '\n' in note.text:
        return f"note: '''{prepare_text_for_dbml(note.text)}'''"
    else:
        return f"note: '{prepare_text_for_dbml(note.text)}'"


def comment_to_dbml(val: str) -> str:
    return comment(val, '//')



from .renderer import DefaultSQLRenderer
from .column import render_column
from .enum import render_enum, render_enum_item
from .expression import render_expression
from .index import render_index
from .note import render_note
from .reference import render_reference
from .table import render_table



import pyparsing as pp

from .common import _
from .common import _c
from .common import c
from .common import end
from .common import n
from .common import note
from .generic import name
from pydbml.parser.blueprints import EnumBlueprint
from pydbml.parser.blueprints import EnumItemBlueprint

pp.ParserElement.set_default_whitespace_chars(' \t\r')

enum_settings = '[' + _ - note('note') + _ - ']' + c


def parse_enum_settings(s, loc, tok):
    '''
    [note: "note content"] // comment
    '''
    result = {}
    if 'note' in tok:
        result['note'] = tok['note']
    if 'comment' in tok:
        result['comment'] = tok['comment'][0]
    return result


enum_settings.set_parse_action(parse_enum_settings)

enum_item = _c + (name('name') + c + enum_settings('settings')[0, 1])


def parse_enum_item(s, loc, tok):
    '''
    student [note: "is stupid"]
    '''
    init_dict = {'name': tok['name']}
    if 'settings' in tok:
        init_dict.update(tok['settings'])
        # comments after settings have priority
        if 'comment' in tok['settings']:
            init_dict['comment'] = tok['settings']['comment']
    if 'comment' not in init_dict and 'comment_before' in tok:
        comment = '\n'.join(c[0] for c in tok['comment_before'])
        init_dict['comment'] = comment

    return EnumItemBlueprint(**init_dict)


enum_item.set_parse_action(parse_enum_item)

enum_body = enum_item[1, ...]

enum_name = pp.Combine(name("schema") + '.' + name("name")) | name("name")

enum = _c + (
    pp.CaselessLiteral('enum')
    - enum_name + _
    - '{'
    + enum_body('items') + n
    - '}'
) + end


def parse_enum(s, loc, tok):
    '''
    enum members {
        janitor
        student
        teacher
        headmaster
    }
    '''
    init_dict = {
        'name': tok['name'],
        'items': list(tok['items'])
    }

    if 'schema' in tok:
        init_dict['schema'] = tok['schema']

    if 'comment_before' in tok:
        comment = '\n'.join(c[0] for c in tok['comment_before'])
        init_dict['comment'] = comment

    return EnumBlueprint(**init_dict)


enum.set_parse_action(parse_enum)



from typing import List
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from .base import SQLObject, DBMLObject
from .column import Column
from .expression import Expression
from .note import Note

if TYPE_CHECKING:  # pragma: no cover
    from .table import Table


class Index(SQLObject, DBMLObject):
    '''Class representing index.'''
    required_attributes = ('subjects', 'table')
    dont_compare_fields = ('table',)

    def __init__(self,
                 subjects: List[Union[str, Column, Expression]],
                 name: Optional[str] = None,
                 unique: bool = False,
                 type: Optional[
                     Literal[
                         # https://www.postgresql.org/docs/current/indexes-types.html
                         "brin",
                         "btree",
                         "gin",
                         "gist",
                         "hash",
                         "spgist",
                     ]
                 ] = None,
                 pk: bool = False,
                 note: Optional[Union[Note, str]] = None,
                 comment: Optional[str] = None):
        self.subjects = subjects
        self.table: Optional[Table] = None

        self.name = name if name else None
        self.unique = unique
        self.type = type
        self.pk = pk
        self.note = Note(note)
        self.comment = comment

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, val: Note) -> None:
        self._note = val
        val.parent = self

    @property
    def subject_names(self):
        '''
        Returns updated list of subject names.
        '''
        return [s.name if isinstance(s, Column) else str(s) for s in self.subjects]

    def __repr__(self):
        '''
        <Index 'test', ['col', '(c*2)']>
        '''

        table_name = self.table.name if self.table else None
        return f"<Index {table_name!r}, {self.subject_names!r}>"

    def __str__(self):
        '''
        Index(test[col, (c*2)])
        '''

        table_name = self.table.name if self.table else ''
        subjects = ', '.join(self.subject_names)
        return f"Index({table_name}[{subjects}])"



from typing import Union

from pydbml.classes import Column, Enum, Expression
from pydbml.renderer.dbml.default.renderer import DefaultDBMLRenderer
from pydbml.renderer.dbml.default.utils import comment_to_dbml, note_option_to_dbml, quote_string, prepare_text_for_dbml
from pydbml.renderer.sql.default.utils import get_full_name_for_sql


def default_to_str(val: Union[Expression, str, int, float]) -> str:
    if isinstance(val, str):
        if val.lower() in ('null', 'true', 'false'):
            return val.lower()
        else:
            return f"'{prepare_text_for_dbml(val)}'"
    elif isinstance(val, Expression):
        return val.dbml
    else:  # int or float or bool
        return str(val)


def render_options(model: Column) -> str:
    options = [ref.dbml for ref in model.get_refs() if ref.inline]
    if model.pk:
        options.append('pk')
    if model.autoinc:
        options.append('increment')
    if model.default:
        options.append(f'default: {default_to_str(model.default)}')
    if model.unique:
        options.append('unique')
    if model.not_null:
        options.append('not null')
    if model.note:
        options.append(note_option_to_dbml(model.note))
    if model.properties:
        if model.table and model.table.database and model.table.database.allow_properties:
            for key, value in model.properties.items():
                options.append(f'{key}: {quote_string(value)}')

    if options:
        return f' [{", ".join(options)}]'
    return ''


@DefaultDBMLRenderer.renderer_for(Column)
def render_column(model: Column) -> str:
    result = comment_to_dbml(model.comment) if model.comment else ''
    result += f'"{model.name}" '
    if isinstance(model.type, Enum):
        result += get_full_name_for_sql(model.type)
    else:
        result += model.type

    result += render_options(model)
    return result



from itertools import chain
from typing import List

from pydbml.classes import Reference, Column
from pydbml.constants import MANY_TO_MANY, MANY_TO_ONE, ONE_TO_ONE, ONE_TO_MANY
from pydbml.exceptions import TableNotFoundError
from pydbml.renderer.sql.default.renderer import DefaultSQLRenderer
from pydbml.renderer.sql.default.utils import comment_to_sql, get_full_name_for_sql


def col_names(cols: List[Column]) -> str:
    return ', '.join(f'"{c.name}"' for c in cols)


def validate_for_sql(model: Reference):
    for col in chain(model.col1, model.col2):
        if col.table is None:
            raise TableNotFoundError(f'Table on {col} is not set')


def generate_inline_sql(model: Reference, source_col: List[Column], ref_col: List[Column]) -> str:
    result = comment_to_sql(model.comment) if model.comment else ''
    result += (
        f'{{c}}FOREIGN KEY ({col_names(source_col)}) '  # type: ignore
        f'REFERENCES {get_full_name_for_sql(ref_col[0].table)} ({col_names(ref_col)})'  # type: ignore
    )
    if model.on_update:
        result += f' ON UPDATE {model.on_update.upper()}'
    if model.on_delete:
        result += f' ON DELETE {model.on_delete.upper()}'
    return result


def generate_not_inline_sql(model: Reference, source_col: List['Column'], ref_col: List['Column']):
    result = comment_to_sql(model.comment) if model.comment else ''
    result += (
        f'ALTER TABLE {get_full_name_for_sql(source_col[0].table)}'  # type: ignore
        f' ADD {{c}}FOREIGN KEY ({col_names(source_col)})'
        f' REFERENCES {get_full_name_for_sql(ref_col[0].table)} ({col_names(ref_col)})' # type: ignore
    )
    if model.on_update:
        result += f' ON UPDATE {model.on_update.upper()}'
    if model.on_delete:
        result += f' ON DELETE {model.on_delete.upper()}'
    return result + ';'


def generate_many_to_many_sql(model: Reference) -> str:
    join_table = model.join_table
    table_sql = join_table.sql  # type: ignore

    n = len(model.col1)
    ref1_sql = generate_not_inline_sql(model, join_table.columns[:n], model.col1)  # type: ignore
    ref2_sql = generate_not_inline_sql(model, join_table.columns[n:], model.col2)  # type: ignore

    result = '\n\n'.join((table_sql, ref1_sql, ref2_sql))
    return result.format(c='')


@DefaultSQLRenderer.renderer_for(Reference)
def render_reference(model: Reference) -> str:
    '''
    Returns SQL of the reference:

    ALTER TABLE "orders" ADD FOREIGN KEY ("customer_id") REFERENCES "customers ("id");

    '''
    validate_for_sql(model)

    if model.type == MANY_TO_MANY:
        return generate_many_to_many_sql(model)

    result = ''
    func = generate_inline_sql if model.inline else generate_not_inline_sql
    if model.type in (MANY_TO_ONE, ONE_TO_ONE):
        result = func(model=model, source_col=model.col1, ref_col=model.col2)
    elif model.type == ONE_TO_MANY:
        result = func(model=model, source_col=model.col2, ref_col=model.col1)

    c = f'CONSTRAINT "{model.name}" ' if model.name else ''

    return result.format(c=c)



import pyparsing as pp

from pydbml.parser.blueprints import TableBlueprint
from .column import table_column, table_column_with_properties
from .common import _, hex_color
from .common import _c
from .common import end
from .common import note
from .common import note_object
from .generic import name, string_literal
from .index import indexes

pp.ParserElement.set_default_whitespace_chars(' \t\r')

alias = pp.WordStart() + pp.Literal('as').suppress() - pp.WordEnd() - name


header_color = (
    pp.CaselessLiteral('headercolor:').suppress() + _
    - pp.Combine(hex_color)('header_color')
)
table_setting = _ + (note('note') | header_color) + _
table_settings = '[' + table_setting + (',' + table_setting)[...] + ']'


def parse_table_settings(s, loc, tok):
    '''
    [headercolor: #cccccc, note: 'note']
    '''
    result = {}
    if 'note' in tok:
        result['note'] = tok['note']
    if 'header_color' in tok:
        result['header_color'] = tok['header_color']
    return result


table_settings.set_parse_action(parse_table_settings)


note_element = note | note_object

prop = name + pp.Suppress(":") + string_literal

table_element = _ + (
    table_column.set_results_name('columns', list_all_matches=True) |
    note_element('note') |
    indexes.set_results_name('indexes', list_all_matches=True)
) + _
table_element_with_property = _ + (
    table_column_with_properties.set_results_name('columns', list_all_matches=True) |
    note_element('note') |
    indexes.set_results_name('indexes', list_all_matches=True) |
    prop.set_results_name('property', list_all_matches=True)
) + _

table_body = table_element[...]
table_body_with_properties = table_element_with_property[...]

table_name = (name('schema') + '.' + name('name')) | (name('name'))

table = _c + (
    pp.CaselessLiteral("table").suppress()
    + table_name
    + alias('alias')[0, 1]
    + table_settings('settings')[0, 1] + _
    + '{' - table_body + _ + '}'
) + end

table_with_properties = _c + (
    pp.CaselessLiteral("table").suppress()
    + table_name
    + alias('alias')[0, 1]
    + table_settings('settings')[0, 1] + _
    + '{' - table_body_with_properties + _ + '}'
) + end


def parse_table(s, loc, tok):
    '''
    Table bookings as bb [headercolor: #cccccc] {
      id integer
      country varchar [NOT NULL, ref: > countries.country_name]
      booking_date date unique pk
      created_at timestamp

      indexes {
          (id, country) [pk] // composite primary key
      }
    }
    '''
    init_dict = {
        'name': tok['name'],
    }
    if 'schema' in tok:
        init_dict['schema'] = tok['schema']
    if 'settings' in tok:
        init_dict.update(tok['settings'])
    if 'alias' in tok:
        init_dict['alias'] = tok['alias'][0]
    if 'note' in tok:
        # will override one from settings
        init_dict['note'] = tok['note'][0]
    if 'indexes' in tok:
        init_dict['indexes'] = tok['indexes'][0]
    if 'columns' in tok:
        init_dict['columns'] = tok['columns']
    if 'comment_before' in tok:
        comment = '\n'.join(c[0] for c in tok['comment_before'])
        init_dict['comment'] = comment
    if 'property' in tok:
        init_dict['properties'] = {k: v for k, v in tok['property']}

    if not init_dict.get('columns'):
        raise SyntaxError(f'Table {init_dict["name"]} at position {loc} has no columns!')

    result = TableBlueprint(**init_dict)

    return result


table.set_parse_action(parse_table)
table_with_properties.set_parse_action(parse_table)



import doctest

from pydbml import database
from pydbml._classes import column
from pydbml._classes import enum
from pydbml._classes import expression
from pydbml._classes import index
from pydbml._classes import note
from pydbml._classes import project
from pydbml._classes import reference
from pydbml._classes import table
from pydbml._classes import table_group
from pydbml.parser import parser


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(column))
    tests.addTests(doctest.DocTestSuite(enum))
    tests.addTests(doctest.DocTestSuite(expression))
    tests.addTests(doctest.DocTestSuite(index))
    tests.addTests(doctest.DocTestSuite(project))
    tests.addTests(doctest.DocTestSuite(note))
    tests.addTests(doctest.DocTestSuite(reference))
    tests.addTests(doctest.DocTestSuite(database))
    tests.addTests(doctest.DocTestSuite(table))
    tests.addTests(doctest.DocTestSuite(table_group))
    tests.addTests(doctest.DocTestSuite(parser))
    return tests



from typing import List
from typing import Optional

from pydbml._classes.base import DBMLObject
from pydbml._classes.note import Note
from pydbml._classes.table import Table


class TableGroup(DBMLObject):
    '''
    TableGroup `items` parameter initially holds just the names of the tables,
    but after parsing the whole document, PyDBMLParseResults class replaces
    them with references to actual tables.
    '''
    dont_compare_fields = ('database',)

    def __init__(self,
                 name: str,
                 items: List[Table],
                 comment: Optional[str] = None,
                 note: Optional[Note] = None,
                 color: Optional[str] = None):
        self.database = None
        self.name = name
        self.items = items
        self.comment = comment
        self.note = note
        self.color = color

    def __repr__(self):
        """
        >>> tg = TableGroup('mygroup', ['t1', 't2'])
        >>> tg
        <TableGroup 'mygroup', ['t1', 't2']>
        >>> t1 = Table('t1')
        >>> t2 = Table('t2')
        >>> tg.items = [t1, t2]
        >>> tg
        <TableGroup 'mygroup', ['t1', 't2']>
        """

        items = [i if isinstance(i, str) else i.name for i in self.items]
        return f'<TableGroup {self.name!r}, {items!r}>'

    def __getitem__(self, key: int) -> Table:
        return self.items[key]

    def __iter__(self):
        return iter(self.items)



from textwrap import indent
from typing import Dict

from pydbml.classes import Project
from pydbml.renderer.dbml.default.renderer import DefaultDBMLRenderer
from pydbml.renderer.dbml.default.utils import comment_to_dbml
from pydbml.tools import doublequote_string


def render_items(items: Dict[str, str]) -> str:
    items_str = ''
    for k, v in items.items():
        if '\n' in v:
            items_str += f"{k}: '''{v}'''\n"
        else:
            items_str += f"{k}: '{v}'\n"
    return indent(items_str.rstrip('\n'), '    ') + '\n'


@DefaultDBMLRenderer.renderer_for(Project)
def render_project(model: Project) -> str:
    result = comment_to_dbml(model.comment) if model.comment else ''
    quoted_name = doublequote_string(model.name)
    result += f'Project {quoted_name} {{\n'
    result += render_items(model.items)
    if model.note:
        result += indent(DefaultDBMLRenderer.render(model.note), '    ') + '\n'
    result += '}'
    return result



"""
These integration tests are based on examples in official DBML docs: https://www.dbml.org/docs/

They should kept up to date to ensure all features are supported by PyDBML.
"""
import os

from pathlib import Path
from unittest import TestCase

from pydbml import PyDBML
from pydbml.classes import Expression, Enum

TEST_DOCS_PATH = Path(os.path.abspath(__file__)).parent / 'test_data/docs'
TestCase.maxDiff = None


class TestDocs(TestCase):
    def test_example(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'example.dbml')
        users, posts = results.tables
        self.assertEqual([c.name for c in users.columns], ['id', 'username', 'role', 'created_at'])
        self.assertEqual([c.name for c in posts.columns], ['id', 'title', 'body', 'user_id', 'created_at'])

        self.assertEqual(len(results.refs), 1)
        ref = results.refs[0]
        self.assertEqual((posts, users), (ref.col1[0].table, ref.col2[0].table))

    def test_project(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'project.dbml')
        proj = results.project
        self.assertEqual(proj.name, 'project_name')
        self.assertEqual(proj.items, {'database_type': 'PostgreSQL'})
        self.assertEqual(proj.note.text, 'Description of the project')

    def test_table_definition(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'table_definition.dbml')
        self.assertEqual(len(results.tables), 1)
        self.assertEqual(results.tables[0].name, 'table_name')

        table = results.tables[0]
        self.assertEqual([c.name for c in table.columns], ['column_name', 'example', 'json_column', 'jsonb_column', 'decimal_column'])

    def test_table_alias(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'table_alias.dbml')
        u, posts = results.tables
        self.assertEqual(u.alias, 'U')
        self.assertIsNone(posts.alias)
        self.assertEqual([c.name for c in u.columns], ['id'])
        self.assertEqual([c.name for c in posts.columns], ['id', 'user_id'])

        self.assertEqual(len(results.refs), 1)
        ref = results.refs[0]
        self.assertEqual((u, posts), (ref.col1[0].table, ref.col2[0].table))

    def test_table_notes(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'table_notes.dbml')
        self.assertEqual(len(results.tables), 1)
        self.assertEqual(results.tables[0].name, 'users')

        table = results.tables[0]
        self.assertEqual([c.name for c in table.columns], ['id', 'status'])
        self.assertEqual(table.note.text, 'Stores user data')
        self.assertEqual(table['status'].note.text, 'status')

    def test_column_settings(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'column_settings.dbml')
        self.assertEqual(len(results.tables), 1)
        self.assertEqual(results.tables[0].name, 'buildings')

        table = results.tables[0]
        self.assertEqual([c.name for c in table.columns], ['address', 'id', 'nullable', 'counter'])
        address, id_, nullable, counter = table.columns
        self.assertTrue(address.unique)
        self.assertTrue(address.not_null)
        self.assertEqual(address.note.text, 'to include unit number')

        self.assertTrue(id_.pk)
        self.assertTrue(id_.unique)
        self.assertEqual(id_.note.text, 'Number')
        self.assertEqual(id_.default, 123)

        self.assertFalse(nullable.not_null)

        self.assertTrue(counter.autoinc)

    def test_default_value(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'default_value.dbml')
        self.assertEqual(len(results.tables), 1)
        self.assertEqual(results.tables[0].name, 'users')

        table = results.tables[0]
        self.assertEqual([c.name for c in table.columns], ['id', 'username', 'full_name', 'gender', 'created_at', 'rating'])
        *_, gender, created_at, rating = table.columns
        self.assertEqual(gender.default, 'm')
        self.assertIsInstance(created_at.default, Expression)
        self.assertEqual(created_at.default.text, 'now()')
        self.assertEqual(rating.default, 10)

    def test_index_definition(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'index_definition.dbml')
        self.assertEqual(len(results.tables), 1)
        self.assertEqual(results.tables[0].name, 'bookings')

        table = results.tables[0]
        self.assertEqual([c.name for c in table.columns], ['id', 'country', 'booking_date', 'created_at'])
        self.assertEqual(len(table.indexes), 8)

        ix = table.indexes
        self.assertEqual(ix[0].subjects, [table['id'], table['country']])
        self.assertEqual(ix[0].comment, 'composite primary key')
        self.assertTrue(ix[0].pk)

        self.assertEqual(ix[1].subjects, [table['created_at']])
        self.assertEqual(ix[1].name, 'created_at_index')
        self.assertEqual(ix[1].note.text, 'Date')

        self.assertEqual(ix[2].subjects, [table['booking_date']])

        self.assertEqual(ix[3].subjects, [table['country'], table['booking_date']])
        self.assertTrue(ix[3].unique)

        self.assertEqual(ix[4].subjects, [table['booking_date']])
        self.assertEqual(ix[4].type, 'hash')

        self.assertEqual(len(ix[5].subjects), 1)
        self.assertIsInstance(ix[5].subjects[0], Expression)
        self.assertEqual(ix[5].subjects[0].text, 'id*2')

        self.assertEqual(len(ix[6].subjects), 2)
        self.assertIsInstance(ix[6].subjects[0], Expression)
        self.assertIsInstance(ix[6].subjects[1], Expression)
        self.assertEqual(ix[6].subjects[0].text, 'id*3')
        self.assertEqual(ix[6].subjects[1].text, 'getdate()')

        self.assertEqual(ix[7].subjects, [Expression('id*3'), table['id']])

    def test_relationships(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'relationships_1.dbml')
        posts, reviews, users = results.tables

        rf = results.refs

        self.assertEqual(rf[0].col1[0].table, posts)
        self.assertEqual(rf[0].col2[0].table, users)
        self.assertEqual(rf[0].type, '>')

        self.assertEqual(rf[1].col1[0].table, reviews)
        self.assertEqual(rf[1].col2[0].table, users)
        self.assertEqual(rf[1].type, '>')

        results = PyDBML.parse_file(TEST_DOCS_PATH / 'relationships_2.dbml')
        posts, reviews, users = results.tables

        rf = results.refs

        self.assertEqual(rf[0].col1[0].table, users)
        self.assertEqual(rf[0].col2[0].table, posts)
        self.assertEqual(rf[0].type, '<')

        self.assertEqual(rf[1].col1[0].table, users)
        self.assertEqual(rf[1].col2[0].table, reviews)
        self.assertEqual(rf[1].type, '<')

    def test_relationships_composite(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'relationships_composite.dbml')
        merchant_periods, merchants = results.tables

        rf = results.refs

        self.assertEqual(len(rf), 1)

        self.assertEqual(rf[0].col1[0].table, merchant_periods)
        self.assertEqual(rf[0].col2[0].table, merchants)
        self.assertEqual(rf[0].type, '>')
        self.assertEqual(
            rf[0].col1,
            [
                merchant_periods['merchant_id'],
                merchant_periods['country_code'],
            ]
        )
        self.assertEqual(
            rf[0].col2,
            [
                merchants['id'],
                merchants['country_code'],
            ]
        )

    def test_relationship_settings(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'relationship_settings.dbml')
        merchant_periods, merchants = results.tables

        rf = results.refs

        self.assertEqual(len(rf), 1)

        self.assertEqual(rf[0].col1[0].table, merchant_periods)
        self.assertEqual(rf[0].col2[0].table, merchants)
        self.assertEqual(rf[0].type, '>')
        self.assertEqual(rf[0].on_delete, 'cascade')
        self.assertEqual(rf[0].on_update, 'no action')

    def test_note_definition(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'note_definition.dbml')
        self.assertEqual(len(results.tables), 1)
        users = results['public.users']
        self.assertEqual(users.note.text, 'This is a note of this table')

    def test_project_notes(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'project_notes.dbml')
        project = results.project

        self.assertEqual(project.name, 'DBML')
        self.assertTrue(project.note.text.startswith('# DBML - Database Markup Language\nDBML'))

    def test_column_notes(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'column_notes.dbml')
        users = results['public.users']

        self.assertEqual(users.note.text, 'Stores user data')
        self.assertEqual(users['column_name'].note.text, 'replace text here')
        self.assertTrue('shipped' in users['status'].note.text)

    def test_enum_definition(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'enum_definition.dbml')
        jobs = results['public.jobs']
        self.assertIsInstance(jobs['status'].type, Enum)
        self.assertIsInstance(jobs['grade'].type, Enum)

        self.assertEqual(len(results.enums), 2)
        js, g = results.enums

        self.assertEqual(js.name, 'job_status')
        self.assertEqual([ei.name for ei in js.items], ['created', 'running', 'done', 'failure'])

        self.assertEqual(g.name, 'grade')
        self.assertEqual([ei.name for ei in g.items], ['A+', 'A', 'A-', 'Not Yet Set'])

    def test_table_group(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'table_group.dbml')

        self.assertEqual(len(results.tables), 5)
        self.assertEqual(len(results.table_groups), 2)

        tb1, tb2, tb3, merchants, countries = results.tables

        tg1, tg2 = results.table_groups

        self.assertEqual(tg1.name, 'tablegroup_name')
        self.assertEqual(tg1.items, [tb1, tb2, tb3])
        self.assertEqual(tg2.name, 'e_commerce1')
        self.assertEqual(tg2.items, [merchants, countries])

    def test_sticky_notes(self) -> None:
        results = PyDBML.parse_file(TEST_DOCS_PATH / 'sticky_notes.dbml')

        self.assertEqual(len(results.sticky_notes), 2)

        sn1, sn2 = results.sticky_notes

        self.assertEqual(sn1.name, 'single_line_note')
        self.assertEqual(sn1.text, 'This is a single line note')
        self.assertEqual(sn2.name, 'multiple_lines_note')
        self.assertEqual(sn2.text, '''This is a multiple lines note\nThis string can spans over multiple lines.''')



import re
from textwrap import indent

from pydbml.classes import Table
from pydbml.renderer.dbml.default.renderer import DefaultDBMLRenderer
from pydbml.renderer.dbml.default.utils import comment_to_dbml, quote_string


def get_full_name_for_dbml(model) -> str:
    if model.schema == 'public':
        return f'"{model.name}"'
    else:
        return f'"{model.schema}"."{model.name}"'


def render_header(model: Table) -> str:
    name = get_full_name_for_dbml(model)

    result = f'Table {name} '
    if model.alias:
        result += f'as "{model.alias}" '
    if model.header_color:
        result += f'[headercolor: {model.header_color}] '
    return result


def render_indexes(model: Table) -> str:
    if model.indexes:
        result = '\n    indexes {\n'
        indexes_str = '\n'.join(DefaultDBMLRenderer.render(i) for i in model.indexes)
        result += indent(indexes_str, '        ') + '\n'
        result += '    }\n'
        return result
    return ''


@DefaultDBMLRenderer.renderer_for(Table)
def render_table(model: Table) -> str:
    result = comment_to_dbml(model.comment) if model.comment else ''
    result += render_header(model)

    result += '{\n'
    columns_str = '\n'.join(DefaultDBMLRenderer.render(c) for c in model.columns)
    result += indent(columns_str, '    ') + '\n'

    if model.properties:
        if model.database and model.database.allow_properties:
            properties_str = '\n' + '\n'.join(f'{key}: {quote_string(value)}' for key, value in model.properties.items()) + '\n'
            properties_str = indent(properties_str, '    ')
            result += properties_str

    if model.note:
        result += indent(model.note.dbml, '    ') + '\n'

    result += render_indexes(model)

    result += '}'
    return result



import pyparsing as pp

from pydbml.parser.blueprints import TableGroupBlueprint, NoteBlueprint
from .common import _, note, note_object, hex_color
from .common import _c
from .common import end
from .generic import name

pp.ParserElement.set_default_whitespace_chars(' \t\r')

table_name = pp.Combine(name + '.' + name) | name
note_element = note | note_object

tg_element = _ + (note_element('note') | table_name.set_results_name('items', list_all_matches=True)) + _

tg_body = tg_element[...]


tg_color = (
    pp.CaselessLiteral('color:').suppress() + _
    - pp.Combine(hex_color)('color')
)
tg_setting = _ + (note('note') | tg_color) + _

tg_settings = '[' + tg_setting + (',' + tg_setting)[...] + ']'

table_group = _c + (
    pp.CaselessLiteral('TableGroup')
    - name('name') + _
    + tg_settings[0, 1] + _
    - '{' + _
    - tg_body + _
    - '}'
) + end


def parse_table_group(s, loc, tok):
    '''
    TableGroup tablegroup_name {
        table1
        table2
        table3
    }
    '''
    init_dict = {
        'name': tok['name'],
        'items': list(tok.get('items', []))
    }
    if 'comment_before' in tok:
        comment = '\n'.join(c[0] for c in tok['comment_before'])
        init_dict['comment'] = comment
    if 'note' in tok:
        note = tok['note']
        init_dict['note'] = note if isinstance(note, NoteBlueprint) else note[0]
    if 'color' in tok:
        init_dict['color'] = tok['color']
    return TableGroupBlueprint(**init_dict)


table_group.set_parse_action(parse_table_group)



from typing import Any

from pydbml.classes import Expression, Index, Column
from pydbml.renderer.sql.default.renderer import DefaultSQLRenderer
from pydbml.renderer.sql.default.utils import comment_to_sql


def render_subject(subject: Any) -> str:
    if isinstance(subject, Column):
        return f'"{subject.name}"'
    elif isinstance(subject, Expression):
        return DefaultSQLRenderer.render(subject)
    else:
        return subject


def render_pk(model: Index, keys: str) -> str:
    result = comment_to_sql(model.comment) if model.comment else ''
    result += f'PRIMARY KEY ({keys})'
    return result


def create_components(model: Index, keys: str) -> str:
    components = []
    if model.comment:
        components.append(comment_to_sql(model.comment))

    components.append('CREATE ')

    if model.unique:
        components.append('UNIQUE ')

    components.append('INDEX ')

    if model.name:
        components.append(f'"{model.name}" ')
    if model.table:
        components.append(f'ON "{model.table.name}" ')

    if model.type:
        components.append(f'USING {model.type.upper()} ')
    components.append(f'({keys})')
    return ''.join(components) + ';'


@DefaultSQLRenderer.renderer_for(Index)
def render_index(model: Index) -> str:
    '''
    Returns inline SQL of the index to be created separately from table
    definition:

    CREATE UNIQUE INDEX ON "products" USING HASH ("id");

    But if it's a (composite) primary key index, returns an inline SQL for
    composite primary key to be used inside table definition:

    PRIMARY KEY ("id", "name")
    '''

    keys = ', '.join(render_subject(s) for s in model.subjects)

    if model.pk:
        return render_pk(model, keys)

    return create_components(model, keys)



from dataclasses import dataclass
from typing import Any
from typing import Collection
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from pydbml.classes import Column
from pydbml.classes import Enum
from pydbml.classes import EnumItem
from pydbml.classes import Expression
from pydbml.classes import Index
from pydbml.classes import Note
from pydbml.classes import Project
from pydbml.classes import Reference
from pydbml.classes import Table
from pydbml.classes import TableGroup
from pydbml._classes.sticky_note import StickyNote
from pydbml.exceptions import ColumnNotFoundError
from pydbml.exceptions import TableNotFoundError
from pydbml.exceptions import ValidationError
from pydbml.tools import remove_indentation
from pydbml.tools import strip_empty_lines


class Blueprint:
    parser = None


@dataclass
class NoteBlueprint(Blueprint):
    text: str

    def _preformat_text(self) -> str:
        '''Preformat the note text for idempotence'''
        result = strip_empty_lines(self.text)
        result = remove_indentation(result)
        return result

    def build(self) -> 'Note':
        text = self._preformat_text()
        return Note(text)


@dataclass
class StickyNoteBlueprint(Blueprint):
    name: str
    text: str

    def _preformat_text(self) -> str:
        '''Preformat the note text for idempotence'''
        result = strip_empty_lines(self.text)
        result = remove_indentation(result)
        return result

    def build(self) -> StickyNote:
        text = self._preformat_text()
        name = self.name
        return StickyNote(name=name, text=text)


@dataclass
class ExpressionBlueprint(Blueprint):
    text: str

    def build(self) -> Expression:
        return Expression(self.text)


@dataclass
class ReferenceBlueprint(Blueprint):
    type: Literal['>', '<', '-', '<>']
    inline: bool
    name: Optional[str] = None
    schema1: str = 'public'
    table1: Optional[str] = None
    col1: Optional[Union[str, Collection[str]]] = None
    schema2: str = 'public'
    table2: Optional[str] = None
    col2: Optional[Union[str, Collection[str]]] = None
    comment: Optional[str] = None
    on_update: Optional[str] = None
    on_delete: Optional[str] = None

    def build(self) -> 'Reference':
        '''
        both tables and columns should be present before build
        '''
        if self.table1 is None:
            raise TableNotFoundError("Can't build Reference, table1 unknown")
        if self.table2 is None:
            raise TableNotFoundError("Can't build Reference, table2 unknown")
        if self.col1 is None:
            raise ColumnNotFoundError("Can't build Reference, col1 unknown")
        if self.col2 is None:
            raise ColumnNotFoundError("Can't build Reference, col2 unknown")

        if self.parser:
            table1 = self.parser.locate_table(self.schema1, self.table1)
        else:
            raise RuntimeError('Parser is not set')

        col1_list = [c.strip('() ') for c in self.col1.split(',')]
        col1 = [table1[col] for col in col1_list]

        table2 = self.parser.locate_table(self.schema2, self.table2)
        col2_list = [c.strip('() ') for c in self.col2.split(',')]
        col2 = [table2[col] for col in col2_list]

        return Reference(
            type=self.type,
            inline=self.inline,
            col1=col1,
            col2=col2,
            name=self.name,
            comment=self.comment,
            on_update=self.on_update,
            on_delete=self.on_delete
        )


@dataclass
class ColumnBlueprint(Blueprint):
    name: str
    type: str
    unique: bool = False
    not_null: bool = False
    pk: bool = False
    autoinc: bool = False
    default: Optional[Any] = None
    note: Optional[NoteBlueprint] = None
    ref_blueprints: Optional[List[ReferenceBlueprint]] = None
    comment: Optional[str] = None
    properties: Optional[Dict[str, str]] = None

    def build(self) -> 'Column':
        if isinstance(self.default, ExpressionBlueprint):
            self.default = self.default.build()
        if self.parser:
            if '.' in self.type:
                schema, name = self.type.split('.')
            else:
                schema, name = 'public', self.type
            for enum in self.parser.database.enums:
                if (enum.schema, enum.name) == (schema, name):
                    self.type = enum
                    break
        return Column(
            name=self.name,
            type=self.type,
            unique=self.unique,
            not_null=self.not_null,
            pk=self.pk,
            autoinc=self.autoinc,
            default=self.default,
            note=self.note.build() if self.note else None,
            comment=self.comment,
            properties=self.properties,
        )


@dataclass
class IndexBlueprint(Blueprint):
    subject_names: List[Union[str, ExpressionBlueprint]]
    name: Optional[str] = None
    unique: bool = False
    type: Optional[
        Literal[
            # https://www.postgresql.org/docs/current/indexes-types.html
            "brin",
            "btree",
            "gin",
            "gist",
            "hash",
            "spgist",
        ]
    ] = None
    pk: bool = False
    note: Optional[NoteBlueprint] = None
    comment: Optional[str] = None

    table = None

    def build(self) -> 'Index':
        return Index(
            # TableBlueprint will process subjects
            subjects=[],
            name=self.name,
            unique=self.unique,
            type=self.type,
            pk=self.pk,
            note=self.note.build() if self.note else None,
            comment=self.comment
        )


@dataclass
class TableBlueprint(Blueprint):
    name: str
    schema: str = 'public'
    columns: Optional[List[ColumnBlueprint]] = None
    indexes: Optional[List[IndexBlueprint]] = None
    alias: Optional[str] = None
    note: Optional[NoteBlueprint] = None
    header_color: Optional[str] = None
    comment: Optional[str] = None
    properties: Optional[Dict[str, str]] = None

    def build(self) -> 'Table':
        result = Table(
            name=self.name,
            schema=self.schema,
            alias=self.alias,
            note=self.note.build() if self.note else None,
            header_color=self.header_color,
            comment=self.comment,
            properties=self.properties
        )
        columns = self.columns or []
        indexes = self.indexes or []
        for col_bp in columns:
            result.add_column(col_bp.build())
        for index_bp in indexes:
            index = index_bp.build()
            new_subjects: List[Union[str, Column, Expression]] = []
            for subj in index_bp.subject_names:
                if isinstance(subj, ExpressionBlueprint):
                    new_subjects.append(subj.build())
                else:
                    for col in result.columns:
                        if col.name == subj:
                            new_subjects.append(col)
                            break
                    else:
                        raise ColumnNotFoundError(
                            f'Cannot add index, column "{subj}" not defined in'
                            ' table "{self.name}".'
                        )
            index.subjects = new_subjects
            result.add_index(index)
        return result

    def get_reference_blueprints(self):
        ''' the inline ones '''
        result = []
        for col in self.columns:
            for ref_bp in col.ref_blueprints or []:
                ref_bp.schema1 = self.schema
                ref_bp.table1 = self.name
                ref_bp.col1 = col.name
                result.append(ref_bp)
        return result


@dataclass
class EnumItemBlueprint(Blueprint):
    name: str
    note: Optional[NoteBlueprint] = None
    comment: Optional[str] = None

    def build(self) -> 'EnumItem':
        return EnumItem(
            name=self.name,
            note=self.note.build() if self.note else None,
            comment=self.comment
        )


@dataclass
class EnumBlueprint(Blueprint):
    name: str
    items: List[EnumItemBlueprint]
    schema: str = 'public'
    comment: Optional[str] = None

    def build(self) -> 'Enum':
        return Enum(
            name=self.name,
            items=[ei.build() for ei in self.items],
            schema=self.schema,
            comment=self.comment
        )


@dataclass
class ProjectBlueprint(Blueprint):
    name: str
    items: Optional[Dict[str, str]] = None
    note: Optional[NoteBlueprint] = None
    comment: Optional[str] = None

    def build(self) -> 'Project':
        return Project(
            name=self.name,
            items=dict(self.items) if self.items else {},
            note=self.note.build() if self.note else None,
            comment=self.comment
        )


@dataclass
class TableGroupBlueprint(Blueprint):
    name: str
    items: List[str]
    comment: Optional[str] = None
    note: Optional[NoteBlueprint] = None
    color: Optional[str] = None

    def build(self) -> 'TableGroup':
        if not self.parser:
            raise RuntimeError('Parser is not set')
        items = []
        for table_name in self.items:
            components = table_name.split('.')
            schema, table = components if len(components) == 2 else ('public', components[0])
            table_obj = self.parser.locate_table(schema, table)
            if table_obj in items:
                raise ValidationError(f'Table "{table}" is already in group "{self.name}"')
            items.append(table_obj)
        return TableGroup(
            name=self.name,
            items=items,
            comment=self.comment,
            note=self.note.build() if self.note else None,
            color=self.color
        )



from itertools import chain
from typing import Collection
from typing import Literal
from typing import Optional
from typing import Union

from pydbml.constants import MANY_TO_MANY
from pydbml.exceptions import DBMLError
from pydbml.exceptions import TableNotFoundError
from .base import SQLObject, DBMLObject
from .column import Column
from .table import Table


class Reference(SQLObject, DBMLObject):
    '''
    Class, representing a foreign key constraint.
    It is a separate object, which is not connected to Table or Column objects
    and its `sql` property contains the ALTER TABLE clause.
    '''
    required_attributes = ('type', 'col1', 'col2')
    dont_compare_fields = ('database', '_inline')

    def __init__(self,
                 type: Literal['>', '<', '-', '<>'],
                 col1: Union[Column, Collection[Column]],
                 col2: Union[Column, Collection[Column]],
                 name: Optional[str] = None,
                 comment: Optional[str] = None,
                 on_update: Optional[str] = None,
                 on_delete: Optional[str] = None,
                 inline: bool = False):
        self.database = None
        self.type = type
        self.col1 = [col1] if isinstance(col1, Column) else list(col1)
        self.col2 = [col2] if isinstance(col2, Column) else list(col2)
        self.name = name if name else None
        self.comment = comment
        self.on_update = on_update
        self.on_delete = on_delete
        self._inline = inline

    @property
    def inline(self) -> bool:
        return self._inline and not self.type == MANY_TO_MANY

    @inline.setter
    def inline(self, val) -> None:
        self._inline = val

    @property
    def join_table(self) -> Optional[Table]:
        if self.type != MANY_TO_MANY:
            return None

        if self.table1 is None:
            raise TableNotFoundError(f"Cannot generate join table for {self}: table 1 is unknown")
        if self.table2 is None:
            raise TableNotFoundError(f"Cannot generate join table for {self}: table 2 is unknown")

        return Table(
            name=f'{self.table1.name}_{self.table2.name}',
            schema=self.table1.schema,
            columns=(
                Column(name=f'{c.table.name}_{c.name}', type=c.type, not_null=True, pk=True)  # type: ignore
                for c in chain(self.col1, self.col2)
            ),
            abstract=True
        )

    @property
    def table1(self) -> Optional[Table]:
        self._validate()
        return self.col1[0].table if self.col1 else None

    @property
    def table2(self) -> Optional[Table]:
        self._validate()
        return self.col2[0].table if self.col2 else None

    def __repr__(self):
        '''
        >>> c1 = Column('c1', 'int')
        >>> c2 = Column('c2', 'int')
        >>> Reference('>', col1=c1, col2=c2)
        <Reference '>', ['c1'], ['c2']>
        >>> c12 = Column('c12', 'int')
        >>> c22 = Column('c22', 'int')
        >>> Reference('<', col1=[c1, c12], col2=(c2, c22))
        <Reference '<', ['c1', 'c12'], ['c2', 'c22']>
        '''

        col1 = ', '.join(f'{c.name!r}' for c in self.col1)
        col2 = ', '.join(f'{c.name!r}' for c in self.col2)
        return f"<Reference {self.type!r}, [{col1}], [{col2}]>"

    def __str__(self):
        '''
        >>> c1 = Column('c1', 'int')
        >>> c2 = Column('c2', 'int')
        >>> print(Reference('>', col1=c1, col2=c2))
        Reference([c1] > [c2]
        >>> c12 = Column('c12', 'int')
        >>> c22 = Column('c22', 'int')
        >>> print(Reference('<', col1=[c1, c12], col2=(c2, c22)))
        Reference([c1, c12] < [c2, c22]
        '''

        col1 = ', '.join(f'{c.name}' for c in self.col1)
        col2 = ', '.join(f'{c.name}' for c in self.col2)
        return f"Reference([{col1}] {self.type} [{col2}]"

    def _validate(self):
        table1 = self.col1[0].table
        if any(c.table != table1 for c in self.col1):
            raise DBMLError('Columns in col1 are from different tables')

        table2 = self.col2[0].table
        if any(c.table != table2 for c in self.col2):
            raise DBMLError('Columns in col2 are from different tables')



from typing import Any

from .base import SQLObject, DBMLObject


class Note(SQLObject, DBMLObject):
    dont_compare_fields = ('parent',)

    def __init__(self, text: Any) -> None:
        self.text: str
        self.text = str(text) if text is not None else ''
        self.parent: Any = None

    def __str__(self):
        '''Note text'''
        return self.text

    def __bool__(self):
        return bool(self.text)

    def __repr__(self):
        '''Note('Note text')'''
        return f'Note({repr(self.text)})'



from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

from .base import SQLObject, DBMLObject
from .note import Note


class EnumItem(SQLObject, DBMLObject):
    '''Single enum item'''

    required_attributes = ('name',)

    def __init__(self,
                 name: str,
                 note: Optional[Union[Note, str]] = None,
                 comment: Optional[str] = None):
        self.name = name
        self.note = Note(note)
        self.comment = comment

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, val: Note) -> None:
        self._note = val
        val.parent = self

    def __repr__(self):
        '''<EnumItem 'en-US'>'''
        return f'<EnumItem {self.name!r}>'

    def __str__(self):
        '''en-US'''
        return self.name


class Enum(SQLObject, DBMLObject):
    required_attributes = ('name', 'schema', 'items')

    def __init__(self,
                 name: str,
                 items: Iterable[Union['EnumItem', str]],
                 schema: str = 'public',
                 comment: Optional[str] = None):
        self.database = None
        self.name = name
        self.schema = schema
        self.comment = comment
        self.items: List[EnumItem] = []
        for item in items:
            self.add_item(item)

    def add_item(self, item: Union['EnumItem', str]) -> None:
        if isinstance(item, EnumItem):
            self.items.append(item)
        elif isinstance(item, str):
            self.items.append(EnumItem(item))

    def __getitem__(self, key: int) -> EnumItem:
        return self.items[key]

    def __iter__(self):
        return iter(self.items)

    def __repr__(self):
        '''
        >>> en = EnumItem('en-US')
        >>> ru = EnumItem('ru-RU')
        >>> Enum('languages', [en, ru])
        <Enum 'languages', ['en-US', 'ru-RU']>
        '''

        item_names = [i.name for i in self.items]
        classname = self.__class__.__name__
        return f'<{classname} {self.name!r}, {item_names!r}>'

    def __str__(self):
        '''
        >>> en = EnumItem('en-US')
        >>> ru = EnumItem('ru-RU')
        >>> print(Enum('languages', [en, ru]))
        languages
        '''

        return self.name



from textwrap import indent

from pydbml.classes import Note
from pydbml.renderer.dbml.default.renderer import DefaultDBMLRenderer
from pydbml.renderer.dbml.default.utils import quote_string


@DefaultDBMLRenderer.renderer_for(Note)
def render_note(model: Note) -> str:
    text = quote_string(model.text)

    text = indent(text, '    ')
    result = f'Note {{\n{text}\n}}'
    return result



from textwrap import indent

from pydbml.classes import Enum, EnumItem
from pydbml.renderer.dbml.default.renderer import DefaultDBMLRenderer
from pydbml.renderer.dbml.default.utils import comment_to_dbml, note_option_to_dbml
from pydbml.renderer.sql.default.utils import get_full_name_for_sql


@DefaultDBMLRenderer.renderer_for(Enum)
def render_enum(model: Enum) -> str:
    result = comment_to_dbml(model.comment) if model.comment else ''
    result += f'Enum {get_full_name_for_sql(model)} {{\n'
    items_str = '\n'.join(DefaultDBMLRenderer.render(i) for i in model.items)
    result += indent(items_str, '    ')
    result += '\n}'
    return result


@DefaultDBMLRenderer.renderer_for(EnumItem)
def render_enum_item(model: EnumItem) -> str:
    result = comment_to_dbml(model.comment) if model.comment else ''
    result += f'"{model.name}"'
    if model.note:
        result += f' [{note_option_to_dbml(model.note)}]'
    return result



import re

from pydbml.classes import Note, Table, Column
from pydbml.renderer.sql.default.renderer import DefaultSQLRenderer


def prepare_text_for_sql(model: Note) -> str:
    '''
    - Process special escape sequence: slash before line break, which means no line break
      https://www.dbml.org/docs/#multi-line-string
    - replace all single quotes with double quotes
    '''

    pattern = re.compile(r'\\\n')
    result = pattern.sub('', model.text)

    result = result.replace("'", '"')
    return result


def generate_comment_on(model: Note, entity: str, name: str) -> str:
    """Generate a COMMENT ON clause out from this note."""
    quoted_text = f"'{prepare_text_for_sql(model)}'"
    note_sql = f'COMMENT ON {entity.upper()} "{name}" IS {quoted_text};'
    return note_sql


@DefaultSQLRenderer.renderer_for(Note)
def render_note(model: Note) -> str:
    """
    For Tables and Columns Note is converted into COMMENT ON clause. All other entities don't
    have notes generated in their SQL code, but as a fallback their notes are rendered as SQL
    comments when sql property is called directly.
    """

    if model.text:
        if isinstance(model.parent, (Table, Column)):
            return generate_comment_on(model, model.parent.__class__.__name__, model.parent.name)
        else:
            text = prepare_text_for_sql(model)
            return '\n'.join(f'-- {line}' for line in text.split('\n'))
    else:
        return ''



from textwrap import indent
from typing import List

from pydbml.constants import MANY_TO_ONE, ONE_TO_ONE, ONE_TO_MANY
from pydbml.classes import Table, Reference, Column
from pydbml.exceptions import UnknownDatabaseError
from pydbml.renderer.sql.default.note import prepare_text_for_sql
from pydbml.renderer.sql.default.renderer import DefaultSQLRenderer
from pydbml.renderer.sql.default.utils import comment_to_sql, get_full_name_for_sql


def get_references_for_sql(model: Table) -> List[Reference]:
    """
    Return all references in the database where this table is on the left side of SQL
    reference definition.
    """
    if not model.database:
        raise UnknownDatabaseError(f'Database for the table {model} is not set')
    result = []
    for ref in model.database.refs:
        if (ref.type in (MANY_TO_ONE, ONE_TO_ONE)) and\
                (ref.table1 == model):
            result.append(ref)
        elif (ref.type == ONE_TO_MANY) and (ref.table2 == model):
            result.append(ref)
    return result


def get_inline_references_for_sql(model: Table) -> List[Reference]:
    '''
    Return inline references for this table sql definition
    '''
    if model.abstract:
        return []
    return [r for r in get_references_for_sql(model) if r.inline]


def create_body(model: Table) -> str:
    body: List[str] = []
    body.extend(indent(DefaultSQLRenderer.render(c), "  ") for c in model.columns)
    body.extend(indent(DefaultSQLRenderer.render(i), "  ") for i in model.indexes if i.pk)
    body.extend(indent(DefaultSQLRenderer.render(r), "  ") for r in get_inline_references_for_sql(model))

    if model._has_composite_pk():
        body.append(
            "  PRIMARY KEY ("
            + ', '.join(f'"{c.name}"' for c in model.columns if c.pk)
            + ')')

    return ',\n'.join(body)


def create_components(model: Table) -> str:
    components = [comment_to_sql(model.comment)] if model.comment else []
    components.append(f'CREATE TABLE {get_full_name_for_sql(model)} (')

    body = create_body(model)

    components.append(body)
    components.append(');')
    components.extend('\n' + DefaultSQLRenderer.render(i) for i in model.indexes if not i.pk)

    return '\n'.join(components)


def render_column_notes(model: Table) -> str:
    result = ''
    for col in model.columns:
        if col.note:
            quoted_note = f"'{prepare_text_for_sql(col.note)}'"
            note_sql = f'COMMENT ON COLUMN "{model.name}"."{col.name}" IS {quoted_note};'
            result += f'\n\n{note_sql}'
    return result


@DefaultSQLRenderer.renderer_for(Table)
def render_table(model: Table) -> str:
    '''
    Returns full SQL for table definition:

    CREATE TABLE "countries" (
      "code" int PRIMARY KEY,
      "name" varchar,
      "continent_name" varchar
    );

    Also returns indexes if they were defined:

    CREATE INDEX ON "products" ("id", "name");
    '''
    result = create_components(model)

    if model.note:
        result += f'\n\n{model.note.sql}'

    result += render_column_notes(model)
    return result



import pyparsing as pp

from .common import _
from .common import _c
from .common import c
from .common import n
from .common import note
from .common import pk
from .common import unique
from .generic import boolean_literal
from .generic import expression
from .generic import expression_literal
from .generic import name
from .generic import number_literal
from .generic import string_literal
from .reference import ref_inline
from pydbml.parser.blueprints import ColumnBlueprint


pp.ParserElement.set_default_whitespace_chars(' \t\r')

type_args = ("(" + pp.original_text_for(expression) + ")")

# column type is parsed as a single string, it will be split by blueprint
column_type = pp.Combine((name + pp.Literal('[]')) | (name + '.' + name) | ((name) + type_args[0, 1]))

default = pp.CaselessLiteral('default:').suppress() + _ - (
    string_literal
    | expression_literal
    | boolean_literal.set_parse_action(
        lambda s, loc, tok: {
            'true': True,
            'false': False,
            'NULL': None
        }[tok[0]]
    )
    | number_literal.set_parse_action(
        lambda s, loc, tok: float(''.join(tok[0])) if '.' in tok[0] else int(tok[0])
    )
)

prop = name + pp.Suppress(":") + string_literal

column_setting = _ + (
    pp.CaselessLiteral("not null").set_parse_action(
        lambda s, loc, tok: True
    )('notnull')
    | pp.CaselessLiteral("null").set_parse_action(
        lambda s, loc, tok: False
    )('notnull')
    | pp.CaselessLiteral("primary key")('pk')
    | pk('pk')
    | unique('unique')
    | pp.CaselessLiteral("increment")('increment')
    | note('note')
    | ref_inline('ref*')
    | default('default')
) + _

column_setting_with_property = column_setting | prop.set_results_name('property', list_all_matches=True)

column_settings = '[' - column_setting + ("," + column_setting)[...] + ']' + c

column_settings_with_properties = '[' - (_ + column_setting_with_property + _) + ("," + column_setting_with_property)[...] + ']' + c


def parse_column_settings(s, loc, tok):
    '''
    [ NOT NULL, increment, default: `now()`]
    '''
    result = {}
    if tok.get('notnull'):
        result['not_null'] = True
    if 'pk' in tok:
        result['pk'] = True
    if 'unique' in tok:
        result['unique'] = True
    if 'increment' in tok:
        result['autoinc'] = True
    if 'note' in tok:
        result['note'] = tok['note']
    if 'default' in tok:
        result['default'] = tok['default'][0]
    if 'ref' in tok:
        result['ref_blueprints'] = list(tok['ref'])
    if 'comment' in tok:
        result['comment'] = tok['comment'][0]
    if 'property' in tok:
        result['properties'] = {k: v for k, v in tok['property']}
    return result


column_settings.set_parse_action(parse_column_settings)
column_settings_with_properties.set_parse_action(parse_column_settings)


constraint = pp.CaselessLiteral("unique") | pp.CaselessLiteral("pk")

table_column = _c + (
    name('name')
    + column_type('type')
    + constraint[...]('constraints') + c
    + column_settings('settings')[0, 1]
) + n


table_column_with_properties = _c + (
    name('name')
    + column_type('type')
    + constraint[...]('constraints') + c
    + column_settings_with_properties('settings')[0, 1]
) + n


def parse_column(s, loc, tok):
    '''
    address varchar(255) [unique, not null, note: 'to include unit number']
    '''
    init_dict = {
        'name': tok['name'],
        'type': tok['type'],
    }
    # deprecated
    for constraint in tok.get('constraints', []):
        if constraint == 'pk':
            init_dict['pk'] = True
        elif constraint == 'unique':
            init_dict['unique'] = True

    if 'settings' in tok:
        init_dict.update(tok['settings'])

    # comments after column definition have priority
    if 'comment' in tok:
        init_dict['comment'] = tok['comment'][0]
    if 'comment' not in init_dict and 'comment_before' in tok:
        comment = '\n'.join(c[0] for c in tok['comment_before'])
        init_dict['comment'] = comment

    return ColumnBlueprint(**init_dict)


table_column.set_parse_action(parse_column)
table_column_with_properties.set_parse_action(parse_column)



import pyparsing as pp

from pydbml.parser.blueprints import ExpressionBlueprint

pp.ParserElement.set_default_whitespace_chars(' \t\r')

name = pp.Word(pp.unicode.alphanums + '_') | pp.QuotedString('"')

# Literals

string_literal = (
    pp.QuotedString("'", escChar="\\")
    ^ pp.QuotedString('"', escChar="\\")
    ^ pp.QuotedString("'''", escChar="\\", multiline=True)
)
expression_literal = pp.Combine(
    pp.Suppress('`')
    + pp.CharsNotIn('`')[...]
    + pp.Suppress('`')
).set_parse_action(lambda s, lok, tok: ExpressionBlueprint(tok[0]))

boolean_literal = (
    pp.CaselessLiteral('true')
    | pp.CaselessLiteral('false')
    | pp.CaselessLiteral('NULL')
)
number_literal = (
    pp.Word(pp.nums)
    ^ pp.Combine(
        pp.Word(pp.nums) + '.' + pp.Word(pp.nums)
    )
)

# Expression

expr_chars = pp.Word(pp.alphanums + "\"'`,._+- \n\t")
expr_chars_no_comma_space = pp.Word(pp.alphanums + "\"'`._+-")
expression = pp.Forward()
factor = (
    pp.Word(pp.alphanums + '_')[0, 1] + '(' + expression + ')'
    | expr_chars_no_comma_space + (pp.Literal(",") | ");" | (pp.LineEnd() + ");"))
    | expr_chars
)
expression << factor[...]



from typing import Dict
from typing import Optional
from typing import Union

from pydbml._classes.base import DBMLObject
from pydbml._classes.note import Note


class Project(DBMLObject):
    dont_compare_fields = ('database',)

    def __init__(self,
                 name: str,
                 items: Optional[Dict[str, str]] = None,
                 note: Optional[Union[Note, str]] = None,
                 comment: Optional[str] = None):
        self.database = None
        self.name = name
        self.items = items or {}
        self.note = Note(note)
        self.comment = comment

    def __repr__(self):
        """<Project 'myproject'>"""
        return f'<Project {self.name!r}>'

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, val: Note) -> None:
        self._note = val
        val.parent = self



from typing import List, Dict
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from pydbml.exceptions import TableNotFoundError
from .base import SQLObject, DBMLObject
from .enum import Enum
from .expression import Expression
from .note import Note

if TYPE_CHECKING:  # pragma: no cover
    from .table import Table
    from .reference import Reference


class Column(SQLObject, DBMLObject):
    '''Class representing table column.'''

    required_attributes = ('name', 'type')
    dont_compare_fields = ('table',)

    def __init__(self,
                 name: str,
                 type: Union[str, Enum],
                 unique: bool = False,
                 not_null: bool = False,
                 pk: bool = False,
                 autoinc: bool = False,
                 default: Optional[Union[str, int, bool, float, Expression]] = None,
                 note: Optional[Union[Note, str]] = None,
                 comment: Optional[str] = None,
                 properties: Union[Dict[str, str], None] = None
                 ):
        self.name = name
        self.type = type
        self.unique = unique
        self.not_null = not_null
        self.pk = pk
        self.autoinc = autoinc
        self.comment = comment
        self.note = Note(note)
        self.properties = properties if properties else {}

        self.default = default
        self.table: Optional['Table'] = None

    def __eq__(self, other: object) -> bool:
        if other is self:
            return True
        if not isinstance(other, self.__class__):
            return False
        self_table = self.table.full_name if self.table else None
        other_table = other.table.full_name if other.table else None
        if self_table != other_table:
            return False
        return super().__eq__(other)

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, val: Note) -> None:
        self._note = val
        val.parent = self

    def get_refs(self) -> List['Reference']:
        '''
        get all references related to this column (where this col is col1 in)
        '''
        if not self.table:
            raise TableNotFoundError('Table for the column is not set')
        return [ref for ref in self.table.get_refs() if self in ref.col1]

    @property
    def database(self):
        return self.table.database if self.table else None

    def __repr__(self):
        '''
        >>> Column('name', 'VARCHAR2')
        <Column 'name', 'VARCHAR2'>
        '''
        type_name = self.type if isinstance(self.type, str) else self.type.name
        return f'<Column {self.name!r}, {type_name!r}>'

    def __str__(self):
        '''
        >>> print(Column('name', 'VARCHAR2'))
        name[VARCHAR2]
        '''

        return f'{self.name}[{self.type}]'



from typing import Iterable, Dict
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from pydbml.exceptions import ColumnNotFoundError
from pydbml.exceptions import IndexNotFoundError
from pydbml.exceptions import UnknownDatabaseError
from .base import SQLObject, DBMLObject
from .column import Column
from .index import Index
from .note import Note

if TYPE_CHECKING:  # pragma: no cover
    from pydbml.database import Database
    from .reference import Reference


class Table(SQLObject, DBMLObject):
    '''Class representing table.'''

    required_attributes = ('name', 'schema')
    dont_compare_fields = ('database',)

    def __init__(self,
                 name: str,
                 schema: str = 'public',
                 alias: Optional[str] = None,
                 columns: Optional[Iterable[Column]] = None,
                 indexes: Optional[Iterable[Index]] = None,
                 note: Optional[Union[Note, str]] = None,
                 header_color: Optional[str] = None,
                 comment: Optional[str] = None,
                 abstract: bool = False,
                 properties: Union[Dict[str, str], None] = None
                 ):
        self.database: Optional[Database] = None
        self.name = name
        self.schema = schema
        self.columns: List[Column] = []
        for column in columns or []:
            self.add_column(column)
        self.indexes: List[Index] = []
        for index in indexes or []:
            self.add_index(index)
        self.alias = alias if alias else None
        self.note = Note(note)
        self.header_color = header_color
        self.comment = comment
        self.abstract = abstract
        self.properties = properties if properties else {}

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, val: Note) -> None:
        self._note = val
        val.parent = self

    @property
    def full_name(self) -> str:
        return f'{self.schema}.{self.name}'

    def _has_composite_pk(self) -> bool:
        return sum(c.pk for c in self.columns) > 1

    def add_column(self, c: Column) -> None:
        '''
        Adds column to self.columns attribute and sets in this column the
        `table` attribute.
        '''
        if not isinstance(c, Column):
            raise TypeError('Columns must be of type Column')
        c.table = self
        self.columns.append(c)

    def delete_column(self, c: Union[Column, int]) -> Column:
        if isinstance(c, Column):
            if c in self.columns:
                c.table = None
                return self.columns.pop(self.columns.index(c))
            else:
                raise ColumnNotFoundError(f'Column {c} if missing in the table')
        elif isinstance(c, int):
            self.columns[c].table = None
            return self.columns.pop(c)

    def add_index(self, i: Index) -> None:
        '''
        Adds index to self.indexes attribute and sets in this index the
        `table` attribute.
        '''
        if not isinstance(i, Index):
            raise TypeError('Indexes must be of type Index')
        for subject in i.subjects:
            if isinstance(subject, Column) and subject.table is not self:
                raise ColumnNotFoundError(f'Column {subject} not in the table')
        i.table = self
        self.indexes.append(i)

    def delete_index(self, i: Union[Index, int]) -> Index:
        if isinstance(i, Index):
            if i in self.indexes:
                i.table = None
                return self.indexes.pop(self.indexes.index(i))
            else:
                raise IndexNotFoundError(f'Index {i} if missing in the table')
        elif isinstance(i, int):
            self.indexes[i].table = None
            return self.indexes.pop(i)

    def get_refs(self) -> List['Reference']:
        if not self.database:
            raise UnknownDatabaseError('Database for the table is not set')
        return [ref for ref in self.database.refs if ref.table1 == self]

    def __getitem__(self, k: Union[int, str]) -> Column:
        if isinstance(k, int):
            return self.columns[k]
        elif isinstance(k, str):
            for c in self.columns:
                if c.name == k:
                    return c
            raise ColumnNotFoundError(f'Column {k} not present in table {self.name}')
        else:
            raise TypeError('indeces must be str or int')

    def get(self, k, default: Optional[Column] = None) -> Optional[Column]:
        try:
            return self.__getitem__(k)
        except (IndexError, ColumnNotFoundError):
            return default

    def __iter__(self):
        return iter(self.columns)

    def __repr__(self):
        '''
        >>> table = Table('customers')
        >>> table
        <Table 'public' 'customers'>
        '''

        return f'<Table {self.schema!r} {self.name!r}>'

    def __str__(self):
        '''
        >>> table = Table('customers')
        >>> table.add_column(Column('id', 'INTEGER'))
        >>> table.add_column(Column('name', 'VARCHAR2'))
        >>> print(table)
        public.customers(id, name)
        '''

        return f'{self.schema}.{self.name}({", ".join(c.name for c in self.columns)})'



import pyparsing as pp

from .generic import string_literal
from pydbml.parser.blueprints import NoteBlueprint

pp.ParserElement.set_default_whitespace_chars(' \t\r')

comment = (
    pp.Suppress("//") + pp.SkipTo(pp.LineEnd())
    | pp.Suppress('/*') + ... + pp.Suppress('*/')
)

# optional comment or newline
_ = ('\n' | comment)[...].suppress()

# optional comment or newline, but comments are captured
_c = (pp.Suppress('\n') | comment('comment_before*'))[...]

# optional captured comment
c = comment('comment')[0, 1]

n = pp.LineEnd()

end = comment[...].suppress() + n | pp.StringEnd()

# obligatory newline
# n = pp.Suppress('\n')[1, ...]

note = pp.CaselessLiteral("note:") + _ - string_literal('text')
note.set_parse_action(lambda s, loc, tok: NoteBlueprint(tok['text']))

note_object = pp.CaselessLiteral('note') + _ - '{' + _ - string_literal('text') + _ - '}'
note_object.set_parse_action(lambda s, loc, tok: NoteBlueprint(tok['text']))

pk = pp.CaselessLiteral("pk")
unique = pp.CaselessLiteral("unique")

hex_char = pp.Word(pp.srange('[0-9a-fA-F]'), exact=1)
hex_color = ("#" - (hex_char * 3 ^ hex_char * 6)).leaveWhitespace()



from typing import List, Any

from pydbml.classes import Index, Expression, Column
from pydbml.renderer.dbml.default.renderer import DefaultDBMLRenderer
from pydbml.renderer.dbml.default.utils import comment_to_dbml, note_option_to_dbml


def render_subjects(source_subjects: List[Any]) -> str:
    subjects = []

    for subj in source_subjects:
        if isinstance(subj, Column):
            subjects.append(subj.name)
        elif isinstance(subj, Expression):
            subjects.append(DefaultDBMLRenderer.render(subj))
        else:
            subjects.append(subj)

    if len(subjects) > 1:
        return f'({", ".join(subj for subj in subjects)})'
    else:
        return subjects[0]


def render_options(model: Index) -> str:
    options = []
    if model.name:
        options.append(f"name: '{model.name}'")
    if model.pk:
        options.append('pk')
    if model.unique:
        options.append('unique')
    if model.type:
        options.append(f'type: {model.type}')
    if model.note:
        options.append(note_option_to_dbml(model.note))

    if options:
        return f' [{", ".join(options)}]'
    return ''


@DefaultDBMLRenderer.renderer_for(Index)
def render_index(model: Index) -> str:
    return (
        (comment_to_dbml(model.comment) if model.comment else '')
        + render_subjects(model.subjects)
        + render_options(model)
    )



from setuptools import setup, find_packages

SHORT_DESCRIPTION = 'Python parser and builder for DBML'

try:
    with open('README.md', encoding='utf8') as readme:
        LONG_DESCRIPTION = readme.read()

except FileNotFoundError:
    LONG_DESCRIPTION = SHORT_DESCRIPTION


setup(
    name='pydbml',
    python_requires='>=3.8',
    description=SHORT_DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    version='1.1.3',
    author='Daniil Minukhin',
    author_email='ddddsa@gmail.com',
    url='https://github.com/Vanderhoof/PyDBML',
    packages=find_packages(exclude=['test', 'test.*']),
    license='MIT',
    platforms='any',
    install_requires=['pyparsing>=3.0.0'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Documentation",
        "Topic :: Text Processing :: Markup",
        "Topic :: Utilities",
    ]
)



import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    pass


def comment(val: str, comb: str) -> str:
    return '\n'.join(f'{comb} {cl}' for cl in val.split('\n')) + '\n'


def indent(val: str, spaces=4) -> str:
    if val == '':
        return val
    return ' ' * spaces + val.replace('\n', '\n' + ' ' * spaces)


def remove_bom(source: str) -> str:
    if source and source[0] == '\ufeff':
        source = source[1:]
    return source


def strip_empty_lines(source: str) -> str:
    """Remove empty lines or lines with just spaces from beginning and end."""
    pattern = re.compile(r'^([ \t]*\n)*(?P<content>[\s\S]+?)(\n[ \t]*)*$')
    return pattern.sub(r'\g<content>', source)


def doublequote_string(source: str) -> str:
    """Safely wrap a single-line string in double quotes"""
    if '\n' in source:
        raise ValueError(f'Multiline strings are not allowed: {source!r}')
    result = source.strip('"').replace('"', '\\"')
    return f'"{result}"'


def remove_indentation(source: str) -> str:
    if not source:
        return source

    pattern = re.compile(r'^\s*')

    lines = source.split('\n')
    spaces = []
    for line in lines:
        if line and not line.isspace():
            indent_match = pattern.search(line)
            if indent_match is not None:  # this is just for you mypy
                spaces.append(len(indent_match[0]))

    indent = min(spaces)
    lines = [l[indent:] for l in lines]
    return '\n'.join(lines)



from unittest import TestCase

import pytest

from pydbml.classes import Note
from pydbml.tools import remove_indentation, doublequote_string
from pydbml.renderer.sql.default.utils import comment_to_sql
from pydbml.tools import indent
from pydbml.renderer.dbml.default.utils import note_option_to_dbml, comment_to_dbml
from pydbml.tools import strip_empty_lines


class TestCommentToDBML(TestCase):
    def test_comment(self) -> None:
        oneline = "comment"
        self.assertEqual(f"// {oneline}\n", comment_to_dbml(oneline))

        expected = """// 
// line1
// line2
// line3
// 
"""
        source = "\nline1\nline2\nline3\n"
        self.assertEqual(comment_to_dbml(source), expected)


class TestCommentToSQL(TestCase):
    def test_comment(self) -> None:
        oneline = "comment"
        self.assertEqual(f"-- {oneline}\n", comment_to_sql(oneline))

        expected = """-- 
-- line1
-- line2
-- line3
-- 
"""
        source = "\nline1\nline2\nline3\n"
        self.assertEqual(comment_to_sql(source), expected)


class TestNoteOptionToDBML(TestCase):
    def test_oneline(self) -> None:
        note = Note("one line note")
        self.assertEqual(f"note: 'one line note'", note_option_to_dbml(note))

    def test_oneline_with_quote(self) -> None:
        note = Note("one line'd note")
        self.assertEqual(f"note: 'one line\\'d note'", note_option_to_dbml(note))

    def test_multiline(self) -> None:
        note = Note("line1\nline2\nline3")
        expected = "note: '''line1\nline2\nline3'''"
        self.assertEqual(expected, note_option_to_dbml(note))

    def test_multiline_with_quotes(self) -> None:
        note = Note("line1\n'''line2\nline3")
        expected = "note: '''line1\n\\'''line2\nline3'''"
        self.assertEqual(expected, note_option_to_dbml(note))


class TestIndent(TestCase):
    def test_empty(self) -> None:
        self.assertEqual(indent(""), "")

    def test_nonempty(self) -> None:
        oneline = "one line text"
        self.assertEqual(indent(oneline), f"    {oneline}")
        source = "line1\nline2\nline3"
        expected = "    line1\n    line2\n    line3"
        self.assertEqual(indent(source), expected)
        expected2 = "  line1\n  line2\n  line3"
        self.assertEqual(indent(source, 2), expected2)


class TestStripEmptyLines(TestCase):
    def test_empty(self) -> None:
        source = ""
        self.assertEqual(strip_empty_lines(source), source)

    def test_no_empty_lines(self) -> None:
        source = "line1\n\n\nline2"
        self.assertEqual(strip_empty_lines(source), source)

    def test_empty_lines(self) -> None:
        stripped = "   line1\n\n line2"
        source = f"\n \n   \n\t \t \n  \n{stripped}\n\n\n   \n \t \n\t \n   \n"
        self.assertEqual(strip_empty_lines(source), stripped)

    def test_one_empty_line(self) -> None:
        stripped = "   line1\n\n line2"
        source = f"\n{stripped}"
        self.assertEqual(strip_empty_lines(source), stripped)
        source = f"{stripped}\n"
        self.assertEqual(strip_empty_lines(source), stripped)

    def test_end(self) -> None:
        stripped = "   line1\n\n line2"
        source = f"\n{stripped}\n   "
        self.assertEqual(strip_empty_lines(source), stripped)


class TestRemoveIndentation(TestCase):
    def test_empty(self) -> None:
        source = ""
        self.assertEqual(remove_indentation(source), source)

    def test_not_empty(self) -> None:
        source = "    line1\n     line2"
        expected = "line1\n line2"
        self.assertEqual(remove_indentation(source), expected)


class TestDoublequoteString:
    @staticmethod
    @pytest.mark.parametrize(
        "source,expected",
        [
            ("Test string", '"Test string"'),
            ('String with "quotes"!', '"String with \\"quotes\\"!"'),
            ('"Quoted string"', '"Quoted string"'),
        ],
    )
    def test_oneline(source: str, expected: str) -> None:
        assert doublequote_string(source) == expected

    @staticmethod
    def test_multiline() -> None:
        with pytest.raises(ValueError):
            doublequote_string('line1\nline2')



import os

from pathlib import Path
from unittest import TestCase

from pydbml import PyDBML
from pydbml.classes import Column
from pydbml.classes import Enum
from pydbml.classes import EnumItem
from pydbml.classes import Expression
from pydbml.classes import Index
from pydbml.classes import Note
from pydbml.classes import Project
from pydbml.classes import Reference
from pydbml.classes import Table
from pydbml.classes import TableGroup
from pydbml.database import Database


TEST_DATA_PATH = Path(os.path.abspath(__file__)).parent / 'test_data'


class TestGenerateDBML(TestCase):
    def create_database(self) -> Database:
        database = Database()
        emp_level = Enum(
            'level',
            [
                EnumItem('junior'),
                EnumItem('middle'),
                EnumItem('senior'),
            ]
        )
        database.add(emp_level)

        t1 = Table('Employees', alias='emp')
        c11 = Column('id', 'integer', pk=True, autoinc=True)
        c12 = Column('name', 'varchar', note=Note('Full employee name'))
        c13 = Column('age', 'number', default=0)
        c14 = Column('level', 'level')
        c15 = Column('favorite_book_id', 'integer')
        t1.add_column(c11)
        t1.add_column(c12)
        t1.add_column(c13)
        t1.add_column(c14)
        t1.add_column(c15)
        database.add(t1)

        t2 = Table('books')
        c21 = Column('id', 'integer', pk=True, autoinc=True)
        c22 = Column('title', 'varchar')
        c23 = Column('author', 'varchar')
        c24 = Column('country_id', 'integer')
        t2.add_column(c21)
        t2.add_column(c22)
        t2.add_column(c23)
        t2.add_column(c24)
        database.add(t2)

        t3 = Table('countries')
        c31 = Column('id', 'integer', pk=True, autoinc=True)
        c32 = Column('name', 'varchar2', unique=True)
        t3.add_column(c31)
        t3.add_column(c32)
        i31 = Index([c32], unique=True)
        t3.add_index(i31)
        i32 = Index([Expression('UPPER(name)')])
        t3.add_index(i32)
        database.add(t3)

        ref1 = Reference('>', c15, c21)
        database.add(ref1)

        ref2 = Reference('<', c31, c24, name='Country Reference', inline=True)
        database.add(ref2)

        tg = TableGroup('Unanimate', [t2, t3])
        database.add(tg)

        p = Project('my project', {'author': 'me', 'reason': 'testing'})
        database.add(p)
        return database

    def test_generate_dbml(self) -> None:
        database = self.create_database()
        with open(TEST_DATA_PATH / 'integration1.dbml') as f:
            expected = f.read()
        self.assertEqual(database.dbml, expected)

    def test_generate_sql(self) -> None:
        database = self.create_database()
        with open(TEST_DATA_PATH / 'integration1.sql') as f:
            expected = f.read()
        self.assertEqual(database.sql, expected)

    def test_parser(self):
        source_path = TEST_DATA_PATH / 'integration1.dbml'
        with self.assertRaises(TypeError):
            PyDBML(2)
        res1 = PyDBML(source_path)
        self.assertIsInstance(res1, Database)
        with open(source_path) as f:
            res2 = PyDBML(f)
        self.assertIsInstance(res2, Database)
        with open(source_path) as f:
            source = f.read()
        res3 = PyDBML(source)
        self.assertIsInstance(res3, Database)
        res4 = PyDBML('\ufeff' + source)
        self.assertIsInstance(res4, Database)

        pydbml = PyDBML()
        self.assertIsInstance(pydbml, PyDBML)
        res5 = pydbml.parse(source)
        self.assertIsInstance(res5, Database)
        res6 = PyDBML.parse('\ufeff' + source)
        self.assertIsInstance(res6, Database)
        res7 = PyDBML.parse_file(str(source_path))
        self.assertIsInstance(res7, Database)
        with open(source_path) as f:
            res8 = PyDBML.parse_file(f)
        self.assertIsInstance(res8, Database)



from typing import Any

from pydbml._classes.base import DBMLObject


class StickyNote(DBMLObject):
    dont_compare_fields = ('database',)

    def __init__(self, name: str, text: Any) -> None:
        self.name = name
        self.text = str(text) if text is not None else ''

        self.database = None

    def __str__(self):
        '''StickyNote('mynote', 'Note text')'''
        return self.__class__.__name__ + f'({repr(self.name)}, {repr(self.text)})'

    def __bool__(self):
        return bool(self.text)

    def __repr__(self):
        '''<StickyNote 'mynote', 'Note text'>'''
        return f'<{self.__class__.__name__} {self.name!r}, {self.text!r}>'



from itertools import chain
from textwrap import indent
from typing import List

from pydbml.classes import Reference, Column
from pydbml.exceptions import TableNotFoundError, DBMLError
from pydbml.renderer.dbml.default.renderer import DefaultDBMLRenderer
from pydbml.renderer.dbml.default.utils import comment_to_dbml
from .table import get_full_name_for_dbml


def validate_for_dbml(model: Reference):
    for col in chain(model.col1, model.col2):
        if col.table is None:
            raise TableNotFoundError(f'Table on {col} is not set')


def render_inline_reference(model: Reference) -> str:
    # settings are ignored for inline ref
    if len(model.col2) > 1:
        raise DBMLError('Cannot render DBML: composite ref cannot be inline')
    table_name = get_full_name_for_dbml(model.col2[0].table)
    return f'ref: {model.type} {table_name}."{model.col2[0].name}"'


def render_col(col: List[Column]) -> str:
    if len(col) == 1:
        return f'"{col[0].name}"'
    else:
        names = (f'"{c.name}"' for c in col)
        return f'({", ".join(names)})'


def render_options(model: Reference) -> str:
    options = []
    if model.on_update:
        options.append(f'update: {model.on_update}')
    if model.on_delete:
        options.append(f'delete: {model.on_delete}')
    if options:
        return f' [{", ".join(options)}]'
    return ''


def render_not_inline_reference(model: Reference) -> str:
    result = comment_to_dbml(model.comment) if model.comment else ''
    result += 'Ref'
    if model.name:
        result += f' {model.name}'

    result += (
        ' {\n    '  # type: ignore
        f'{get_full_name_for_dbml(model.table1)}.{render_col(model.col1)} '
        f'{model.type} '
        f'{get_full_name_for_dbml(model.table2)}.{render_col(model.col2)}'
        f'{render_options(model)}'
        '\n}'
    )
    return result


@DefaultDBMLRenderer.renderer_for(Reference)
def render_reference(model: Reference) -> str:
    validate_for_dbml(model)
    if model.inline:
        return render_inline_reference(model)
    else:
        return render_not_inline_reference(model)



import os

from pathlib import Path
from unittest import TestCase

from pyparsing import ParserElement

from pydbml import PyDBML


ParserElement.set_default_whitespace_chars(' \t\r')


TEST_DATA_PATH = Path(os.path.abspath(__file__)).parent / 'test_data'


class EditingTestCase(TestCase):
    def setUp(self):
        self.dbml = PyDBML(TEST_DATA_PATH / 'editing.dbml')


class TestEditTable(EditingTestCase):
    def test_name(self) -> None:
        products = self.dbml['public.products']
        products.name = 'changed_products'
        self.assertIn('CREATE TABLE "changed_products"', products.sql)
        self.assertIn('CREATE INDEX "product_status" ON "changed_products"', products.sql)
        self.assertIn('Table "changed_products"', products.dbml)

        ref = self.dbml.refs[0]
        self.assertIn('ALTER TABLE "changed_products"', ref.sql)
        self.assertIn('"changed_products"."merchant_id"', ref.dbml)

        index = products.indexes[0]
        self.assertIn('ON "changed_products"', index.sql)

    def test_alias(self) -> None:
        products = self.dbml['public.products']
        products.alias = 'new_alias'

        self.assertIn('as "new_alias"', products.dbml)


class TestColumn(EditingTestCase):
    def test_name(self) -> None:
        products = self.dbml['public.products']
        col = products['name']
        col.name = 'new_name'
        self.assertEqual(col.sql, '"new_name" varchar')
        self.assertEqual(col.dbml, '"new_name" varchar')
        self.assertIn('"new_name" varchar', products.sql)
        self.assertIn('"new_name" varchar', products.dbml)

        self.assertEqual(col, products[col.name])

    def test_name_index(self) -> None:
        products = self.dbml['public.products']
        col = products['status']
        col.name = 'changed_status'
        self.assertIn('"changed_status"', products.indexes[0].sql)
        self.assertIn('changed_status', products.indexes[0].dbml)
        self.assertIn(
            'CREATE INDEX "product_status" ON "products" ("merchant_id", "changed_status");',
            products.sql
        )
        self.assertIn(
            "(merchant_id, changed_status) [name: 'product_status']",
            products.dbml
        )

    def test_name_ref(self) -> None:
        products = self.dbml['public.products']
        col = products['merchant_id']
        col.name = 'changed_merchant_id'
        merchants = self.dbml['public.merchants']
        table_ref = merchants.get_refs()[0]
        self.assertIn('FOREIGN KEY ("changed_merchant_id")', table_ref.sql)


class TestEnum(EditingTestCase):
    def test_enum_name(self):
        products = self.dbml['public.products']
        enum = self.dbml.enums[0]
        enum.name = 'changed product status'
        self.assertIn('CREATE TYPE "changed product status"', enum.sql)
        self.assertIn('Enum "changed product status"', enum.dbml)

        col = products['status']
        self.assertEqual(col.sql, '"status" "changed product status"')
        self.assertEqual(col.dbml, '"status" "changed product status"')
